/******************************************************************************/
/*                                                                            */
/*  File: swf.cpp                                                             */
/*  Author: bkenwright@xbdev.net                                              */
/*  Date: 19-01-06                                                            */
/*                                                                            */
/******************************************************************************/
/*
    swf file format loader/parser

	see .h file for further details
*/
/******************************************************************************/

#include <windows.h>	// HMODULE, and loading dll
#include <stdio.h>
#include <stdarg.h>
#include <string>

#include "loadjpg.h"	// DecodeJpgFileData(..)

#include "swf.h"

//------------------------------------------------------------------------------

//Saving debug information to a log file
__forceinline void dprintf(const char *fmt, ...) 
{
	/*
	va_list parms;
	char buf[256];

	// Try to print in the allocated space.
	va_start(parms, fmt);
	vsprintf (buf, fmt, parms);
	va_end(parms);

	// Write the information out to a txt file
	FILE *fp = fopen("output.txt", "a+");
	fprintf(fp, "%s",  buf);
	fclose(fp);
	*/


}// End dprintf(..)


int Negate(int val, int numbits)
{
	bool neg = ((val >> (numbits-1)) & 0x1);
	if (neg)
	{
		// add extra 1's to make it negative
		for (int i=numbits-1; i<32; i++)
		{
			val |= (1<<i);
		}
		val = (~val)+1;
		val = -val;
	}
	return val;
}


//------------------------------------------------------------------------------

int ReadBitData(char * uncompData, int& val, int numBitsToRead, int & bitOffset)
{
	// Determine how many bytes this read will use
	int numBytes = 0;

	int buf = 0;
	int numBits = 8 - bitOffset;

	int data = uncompData[0] << bitOffset;

	int i=0;

	buf = 0;
	for (int j=0; j<numBitsToRead; j++)
	{
		if (data & 0x80)
		{
			buf = buf | 1;		// We have a 1 at this bit
		}
		buf = buf << 1;			// Defaults to 0, so we have 0 at this bit
		data = data<<1;

		numBits--;
		if (numBits==0)
		{
			uncompData++;
			data = *uncompData;
			numBits=8;

			numBytes++;			// Total number of bytes we've read
		}
	}
	buf = buf>>1;				// Took me a while to track this down, but we need

	val = buf;

	bitOffset = 8 - numBits;

	return numBytes;
}


//------------------------------------------------------------------------------
/*
stDefineShape::stDefineShape() : stSWFBase(TagID_DEFINESHAPE2)
{
	m_tags=NULL;
}
*/

int ReadMatrix(char * uncompData, stMatrix* matrix, bool doPad)
{
	int val = 0;
	int numBitsToRead = 0;
	int bitOffset = 0;
	float result = 0.0f;
	char* startAddr = uncompData;

	dprintf("\tMatrix:\n");

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	matrix->m_hasScale = val;
	dprintf("\t hasScale: %d\n", val);

	if (matrix->m_hasScale > 0 )
	{
		int numScaleBits = 0;
		uncompData += ReadBitData(uncompData, numScaleBits, 5, bitOffset);
		matrix->m_numScaleBits = numScaleBits;
		dprintf("\t numScaleBits: %d\n", numScaleBits);

		uncompData += ReadBitData(uncompData, val, numScaleBits, bitOffset);
		result = ((static_cast<int>(val)) / 65536.0f);
		matrix->m_xScale = result;
		dprintf("\t xScale: %.1f\n", result);

		uncompData += ReadBitData(uncompData, val, numScaleBits, bitOffset);
		result = ((static_cast<int>(val)) / 65536.0f);
		matrix->m_yScale = result;
		dprintf("\t yScale: %.1f\n", result);
	}

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	matrix->m_hasRotate = val;

	if (matrix->m_hasRotate)
	{
		int numRotateBits = 0;
		uncompData += ReadBitData(uncompData, numRotateBits, 5, bitOffset);
		matrix->m_numRotateBits = numRotateBits;
		dprintf("\t numScaleBits: %d\n", numRotateBits);

		uncompData += ReadBitData(uncompData, val, numRotateBits, bitOffset);
		val = Negate(val, numRotateBits); 
		result = ((static_cast<int>(val)) / 65536.0f);
		matrix->m_RotateSkew0 = result;
		dprintf("\t RotateSkew0: %.1f\n", result);

		uncompData += ReadBitData(uncompData, val, numRotateBits, bitOffset);
		val = Negate(val, numRotateBits);
		result = ((static_cast<int>(val)) / 65536.0f);
		matrix->m_RotateSkew1 = result;
		dprintf("\t RotateSkew0: %.1f\n", result);
	}



	int numTranslateBits = 0;
	uncompData += ReadBitData(uncompData, numTranslateBits, 5, bitOffset);
	matrix->m_numTranslateBits = numTranslateBits;
	dprintf("\t numTranslateBits: %d\n", numTranslateBits);

	uncompData += ReadBitData(uncompData, val, numTranslateBits, bitOffset);
	val = Negate(val, numTranslateBits);
	matrix->m_xTranslate = val;
	dprintf("\t xTranslate: %d\n", val);

	uncompData += ReadBitData(uncompData, val, numTranslateBits, bitOffset);
	val = Negate(val, numTranslateBits);
	matrix->m_yTranslate = val;
	dprintf("\t yTranslate: %d\n", val);

	if (doPad)
	{
		if (bitOffset>0)
		{
			dprintf("\t Padding\n");
			uncompData+=1;
		}
	}
	

	int bytesRead = (int)(uncompData - startAddr);

	return bytesRead;
}

//------------------------------------------------------------------------------

int ReadCXForm(char * uncompData, stCXForm* out)
{
	int val = 0;
	int numBitsToRead = 0;
	int bitOffset = 0;
	float result = 0.0f;
	char* startAddr = uncompData;

	dprintf("\tColorTransform CXForm:\n");

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_HasAddTerms = val;
	dprintf("\t HasAddTerms: %d\n", val);

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_HasMultTerms = val;
	dprintf("\t HasMultTerms: %d\n", val);

	uncompData += ReadBitData(uncompData, val, 4, bitOffset);
	out->m_Nbits = val;
	dprintf("\t Nbits: %d\n", val);

	int nBits = out->m_Nbits;

	if (out->m_HasMultTerms)
	{
		int r, g, b, a;
		uncompData += ReadBitData(uncompData, r, nBits, bitOffset);
		uncompData += ReadBitData(uncompData, g, nBits, bitOffset);
		uncompData += ReadBitData(uncompData, b, nBits, bitOffset);
		uncompData += ReadBitData(uncompData, a, nBits, bitOffset);

		float div = 256.0f;
		out->m_RedMultTerm    = (float)r/div;
		out->m_GreenMultTerm  = (float)g/div;
		out->m_BlueMultTerm   = (float)b/div;
		out->m_AlphaMultTerm  = (float)a/div;

		out->m_RedMultTerm   = max(0, min(out->m_RedMultTerm,1.0f));
		out->m_GreenMultTerm = max(0, min(out->m_GreenMultTerm,1.0f));
		out->m_BlueMultTerm  = max(0, min(out->m_BlueMultTerm,1.0f));
		out->m_AlphaMultTerm = max(0, min(out->m_AlphaMultTerm,1.0f));
	}

	if (out->m_HasAddTerms)
	{
		uncompData += ReadBitData(uncompData, out->m_RedAddTerm,   nBits, bitOffset);
		uncompData += ReadBitData(uncompData, out->m_GreenAddTerm, nBits, bitOffset);
		uncompData += ReadBitData(uncompData, out->m_BlueAddTerm,  nBits, bitOffset);
		uncompData += ReadBitData(uncompData, out->m_AlphaAddTerm, nBits, bitOffset);

		out->m_RedAddTerm   = Negate(out->m_RedAddTerm, nBits);
		out->m_GreenAddTerm = Negate(out->m_GreenAddTerm, nBits);
		out->m_BlueAddTerm  = Negate(out->m_BlueAddTerm, nBits);
		out->m_AlphaAddTerm = Negate(out->m_AlphaAddTerm, nBits);

	}

	if (bitOffset>0)
	{
		dprintf("\t Padding\n");
		uncompData+=1;
	}

	int bytesRead = (int)(uncompData - startAddr);
	return bytesRead;
}

//------------------------------------------------------------------------------

int ReadRect(char * uncompData, stRect* rect)
{
	dprintf("stRect:\n");

	// Determine how many bytes Rect uses
	int numBytes = 0;

	unsigned char rectFirstByte = *uncompData;
	uncompData++;
	numBytes+=1;

	int nbits = rectFirstByte >> 3;
	rect->m_Nbits = nbits;

	dprintf(" +- nbits: %d\n", nbits);

	int buf = 0;
	int numBits = 3;
	int data = rectFirstByte & 0x7;

	int i=0;
	while (i<4)
	{
		buf = 0;
		for (int j=0; j<nbits; j++)
		{
			if (data & 0x80)
			{
				buf = buf | 1;		// We have a 1 at this bit
			}
			buf = buf << 1;			// Defaults to 0, so we have 0 at this bit
			data = data<<1;

			numBits--;
			if (numBits==0)
			{
				data = *uncompData;
				uncompData++;
				numBits=8;

				//tmpOutLen--;		// Total bytes left in our data
				numBytes++;
			}
		}
		buf = buf>>1;				// Took me a while to track this down, but we need
		// to go back one as we have found our value
		dprintf(" +- buf: %d\n", buf);																			

		rect->m_sizes[i] = buf;

		i++;
	}

	return numBytes;
}

//------------------------------------------------------------------------------

int ReadBackGroundColour(char * uncompData, stBackGroundColour* out)
{
	out->m_R = uncompData[0];
	out->m_G = uncompData[1];
	out->m_B = uncompData[2];

	dprintf(" +- stBackGroundColour : RGB: 0x%.2X, 0x%.2X, 0x%.2X\n", out->m_R, out->m_G, out->m_B);
	return 3;
}

//------------------------------------------------------------------------------

int ReadFrameLabel(char * uncompData, stFrameLabel* out)
{
	memset(out->m_name, 0, sizeof(out->m_name));
	strcpy(out->m_name, uncompData);

	dprintf(" +- stFrameLabel : Name: %s\n", out->m_name);
	return 0;
}

//------------------------------------------------------------------------------

void DebugRawDataFile(char* filename, char * data, int size)
{
	FILE *fp = fopen(filename, "wb+");
	fwrite(data, size, 1, fp);
	fclose(fp);
}

//------------------------------------------------------------------------------


int ReadDefineBitsJPEG2(char * uncompData, stDefineBitsJPEG2* out, int tagLen)
{
	out->m_characterID = *((unsigned short*)&uncompData[0]);
	uncompData += 2;
	tagLen -= 2;

	out->m_jpgData = uncompData;


	dprintf("\tstDefineBitsJPEG2 : CharacterID: %d\n", out->m_characterID);

	static int count = 0;
	count++;
	char buf[32];
	sprintf(buf, "imageJPEG2_%d.jpg", count);

	// ??? Not sure what this is about!  Should start with the SOI jpg tag, but
	// instead seems to start with 4 misc bytes...looks like EOI and SOI tags,
	// back to front!  But why would that be?
	uncompData+=4;
	tagLen -= 4;

	{
		//dprintf("Dumping To File %s\n", buf);
		//DebugRawDataFile(buf, &uncompData[0], tagLen);


		// Decompressing the jpg image data and write it to a bmp
		const unsigned char* jpgRawData = (const unsigned char*)&uncompData[0];
		unsigned char* rgbpix = NULL;
		unsigned int width = 0;
		unsigned int height = 0;
		DecodeJpgFileData(jpgRawData, tagLen, &rgbpix, &width, &height);

		// Dumping it to a bmp file
		//sprintf(buf, "imageJPEG2_to_BMP%d.bmp", count);
		//WriteBMP24(buf, width, height, rgbpix); 
		//delete[] rgbpix;

		char* rgb = (char*)rgbpix;
		out->m_width = width;
		out->m_height = height;


		// Create a single array of RGBA, instead of rgb and alpha separate
		out->m_rgba = new char[ width * height * 4 ];
		unsigned int* rgba = (unsigned int*)out->m_rgba;
		for (int y=0; y<(int)height; y++)
		{
			for (int x=0; x<(int)width; x++)
			{
				int indx = x + y*width;
				unsigned char r = rgb[indx*3 + 0];
				unsigned char g = rgb[indx*3 + 1];
				unsigned char b = rgb[indx*3 + 2];
				unsigned char a = 0xff;

				rgba[indx] = (r<<24)|(g<<16)|(b<<8)|(a);
			}
		}

		delete[] rgbpix; // decoded jmp data
		rgbpix=NULL;
	}

	return 0;
}

//------------------------------------------------------------------------------

int ReadDefineBitsJPEG3(char * uncompData, stDefineBitsJPEG3* out, int tagLen)
{
	out->m_characterID = *((unsigned short*)&uncompData[0]);
	uncompData += 2;
	tagLen -= 2;

	out->m_sizeJpgData = *((unsigned int*)&uncompData[0]);
	uncompData += 4;
	tagLen -= 4;

	out->m_jpgData = uncompData;


	dprintf("\tstDefineBitsJPEG3 : CharacterID: %d\n", out->m_characterID);

	static int count = 0;
	count++;
	char buf[32];
	sprintf(buf, "imageJPEG3_%d.jpg", count);

	{
		//dprintf("Dumping To File %s\n", buf);
		//DebugRawDataFile(buf, &uncompData[0], out->m_sizeJpgData);

		// Decompressing the jpg image data and write it to a bmp
		const unsigned char* jpgRawData = (const unsigned char*)&uncompData[0];
		unsigned char* rgbpix = NULL;
		unsigned int width = 0;
		unsigned int height = 0;
		DecodeJpgFileData(jpgRawData, out->m_sizeJpgData, &rgbpix, &width, &height);

		// Dumping it to a bmp file
		//sprintf(buf, "imageJPEG3_to_BMP%d.bmp", count);
		//WriteBMP24(buf, width, height, rgbpix); 
		//delete[] rgbpix;

		char* rgb = (char*)rgbpix;
		out->m_width = width;
		out->m_height = height;

		// Lets do the alpha part on the end....

		// Increment past the jpg data to the alpha part
		int alphaSize = out->m_sizeJpgData;
		char* alphaData = uncompData + out->m_sizeJpgData;

		// -- Alpha part of grabbing the data ----

		// The jpg3 alpha component is zlib compressed, so we have to use the zlib
		// decompression algorithm again
		int (*uncompress) (char *dest, unsigned int *destLen, const char *source, unsigned int sourceLen);
		HMODULE dll = LoadLibrary("zlib1.dll");
		uncompress = (int (*) (char *dest, unsigned int *destLen, const char *source, unsigned int sourceLen))GetProcAddress(dll, "uncompress");


		// Just allocate a big junk of memory, read it all in, and uncompress it!
		int outSize = 500000;
		// This is where our uncompressed data will go
		unsigned char * tmpOut = new unsigned char [outSize];
		memset(tmpOut, 0, outSize);
		// We have to set tmpOutLen to a val, as it modifies it to the actual size of
		// the decompressed data
		unsigned int tmpOutLen = outSize;
		// go go go...do the uncompression
		int error = uncompress((char*)tmpOut, &tmpOutLen, alphaData, alphaSize);


		// Just for info, the zlib error codes
		//#define Z_OK 0
		//#define Z_STREAM_END 1
		//#define Z_NEED_DICT 2
		//#define Z_ERRNO (-1)
		//#define Z_STREAM_ERROR (-2)
		//#define Z_DATA_ERROR (-3)
		//#define Z_MEM_ERROR (-4)
		//#define Z_BUF_ERROR (-5)
		//#define Z_VERSION_ERROR (-6)

		if (error!=0)
		{
			dprintf("Error decompressing JPEG3 Alpha - zlib error: %d", error);
		}

		// Create a single array of RGBA, instead of rgb and alpha separate
		out->m_rgba = new char[ width * height * 4 ];
		unsigned int* rgba = (unsigned int*)out->m_rgba;
		for (int y=0; y<(int)height; y++)
		{
			for (int x=0; x<(int)width; x++)
			{
				int indx = x + y*width;
				unsigned char r = rgb[indx*3 + 0];
				unsigned char g = rgb[indx*3 + 1];
				unsigned char b = rgb[indx*3 + 2];
				unsigned char a = tmpOut[indx];

				rgba[indx] = (r<<24)|(g<<16)|(b<<8)|(a);
			}
		}

		delete[] tmpOut; // decoded alpha data
		tmpOut=NULL;

		delete[] rgbpix; // decoded jmp data
		rgbpix=NULL;

	}
	return 0;
}

//------------------------------------------------------------------------------

int ReadGradientRecord(char * uncompData, stGradientRecord* out, enShapeType shapeType)
{
	char* startAddr = uncompData;
	out->m_Ratio = uncompData[0];
	uncompData++;

	int R = uncompData[0];
	int G = uncompData[1];
	int B = uncompData[2];
	int A = 0xff;

	switch (shapeType)
	{
		case SHAPE1:
		case SHAPE2:
		{
			uncompData+=3;
		}
		break;

		case SHAPE3:
		{
			A = uncompData[3];
			uncompData+=4;
		}
		break;
	}

	out->m_Colour = ((R<<24)|(G<<16)|(B<<8)|A);
	
	int numBytesRead = (int)(uncompData - startAddr);
	return numBytesRead;
}

//------------------------------------------------------------------------------

int ReadGradient(char * uncompData, stGradient* out, enShapeType shapeType)
{
	char* startAddr = uncompData;

	out->m_NumGradients = uncompData[0];
	uncompData+=1;
	out->m_GradientRecords = new stGradientRecord[out->m_NumGradients];

	for (int i=0; i<out->m_NumGradients; i++)
	{
		uncompData += ReadGradientRecord(uncompData, &out->m_GradientRecords[i], shapeType);
	}


	int numBytesRead = (int)(uncompData - startAddr);

	return numBytesRead;
}

//------------------------------------------------------------------------------

int ReadFillStyle(char * uncompData, stFillStyle* fs, enShapeType shapeType)
{
	char* startAddr = uncompData;


	fs->m_FillStyleType = (enFillStyleType)uncompData[0];
	uncompData++;

	switch (fs->m_FillStyleType)
	{
		case FS_SOLID:
		{
			int R = uncompData[0];
			int G = uncompData[1];
			int B = uncompData[2];
			int A = 0xff;

			switch (shapeType)
			{
				case SHAPE1:
				case SHAPE2:
					uncompData+=3;
				break;
				
				case SHAPE3:
				{
					A = uncompData[3];
					uncompData+=4;
				}
				break;
			}

			fs->m_RGBA = ((R<<24)|(G<<16)|(B<<8)|A);
		}
		break;

		case FS_LINEAR:
		case FS_RAIAL:
		{
			stMatrix* mat = &fs->m_GradientMatrix;
			uncompData += ReadMatrix(uncompData, mat);

			stGradient* grad = &fs->m_Gradient;
			uncompData += ReadGradient(uncompData, grad, shapeType);
		}
		break;

		case FS_TILED:
		case FS_CLIPPED:
		{
			fs->m_BitmapID = *((unsigned short*)(uncompData));
			uncompData+=2;

			stMatrix* mat = &fs->m_BitmapMatrix;
			uncompData += ReadMatrix(uncompData, mat);
		}
		break;

	}

	int numBytesRead = (int)(uncompData - startAddr);

	return numBytesRead;
}

//------------------------------------------------------------------------------


int ReadLineStyle(char * uncompData, stLineStyle* ls, enShapeType shapeType)
{
	char* startAddr = uncompData;

	unsigned short width = ((unsigned short*)(uncompData))[0];
	uncompData+=2;
	ls->m_Width = width;

	switch (shapeType)
	{
		case SHAPE1:
		case SHAPE2:
		{
			int R = uncompData[0];
			int G = uncompData[1];
			int B = uncompData[2];
			uncompData+=3;
			ls->m_RGBA = (R<<24) | (G<<16) | (B<<8) | (0xFF);


			break;
		}
		case SHAPE3:
		{
			int R = uncompData[0];
			int G = uncompData[1];
			int B = uncompData[2];
			int A = uncompData[3];
			uncompData+=4;
			ls->m_RGBA = (R<<24) | (G<<16) | (B<<8) | (A);

			break;
		}
	};


	int numBytesRead = (int)(uncompData - startAddr);

	return numBytesRead;
}

//------------------------------------------------------------------------------


int ReadShapeRecordType1(char* uncompData, stShapeWithStyle* sws, stShapeRecordType1* out, int& bitOffset, enShapeType shapeType)
{
	int val = 0;
	int numBitsToRead = 0;
	float result = 0.0f;
	char* startAddr = uncompData;

	dprintf("\tShapeRecordType1:\n");

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	dprintf("\t TypeFlag: %d\n", val);
	out->m_TypeFlag=val;

	uncompData += ReadBitData(uncompData, out->m_StateNewStyles, 1, bitOffset);
	dprintf("\t StateNewStyles: %d\n", out->m_StateNewStyles);

	uncompData += ReadBitData(uncompData, out->m_StateLineStyle, 1, bitOffset);
	dprintf("\t StateLineStyle: %d\n", out->m_StateLineStyle);
	

	uncompData += ReadBitData(uncompData, out->m_StateFillStyle1, 1, bitOffset);
	dprintf("\t StateFillStyle1: %d\n", out->m_StateFillStyle1);
	
	uncompData += ReadBitData(uncompData, out->m_StateFillStyle0, 1, bitOffset);
	dprintf("\t StateFillStyle0: %d\n", out->m_StateFillStyle0);

	
	uncompData += ReadBitData(uncompData, out->m_StateMoveTo, 1, bitOffset);
	dprintf("\t StateMoveTo: %d\n", out->m_StateMoveTo);

	if (out->m_StateMoveTo)
	{
		int MoveBits = 0;
		uncompData += ReadBitData(uncompData, MoveBits, 5, bitOffset);
		dprintf("\t MoveBits: %d\n", MoveBits);
		out->m_MoveBits = MoveBits;

		
		uncompData += ReadBitData(uncompData, out->m_MoveX, MoveBits, bitOffset);
		dprintf("\t MoveX: %d\n", out->m_MoveX);

		bool neg = ((out->m_MoveX >> (MoveBits-1)) & 0x1);
		if (neg)
		{
			// add extra 1's to make it negative
			for (int i=MoveBits-1; i<32; i++)
			{
				out->m_MoveX |= (1<<i);
			}
			out->m_MoveX = (~out->m_MoveX)+1;
			out->m_MoveX = -out->m_MoveX;
		}

		uncompData += ReadBitData(uncompData, out->m_MoveY, MoveBits, bitOffset);
		
		neg = ((out->m_MoveY >> (MoveBits-1)) & 0x1);
		if (neg)
		{
			// add extra 1's to make it negative
			for (int i=MoveBits-1; i<32; i++)
			{
				out->m_MoveY |= (1<<i);
			}
			out->m_MoveY = (~out->m_MoveY)+1;
			out->m_MoveY = -out->m_MoveY;
		}


		dprintf("\t MoveY: %d\n", out->m_MoveY);
	}

	int numFillBits = sws->m_NumFillBits;
	int numLineBits = sws->m_NumLineBits;

	if (out->m_StateFillStyle0)
	{
		uncompData += ReadBitData(uncompData, out->m_Fill0Style, numFillBits, bitOffset);
		dprintf("\t Fill0Style: %d\n", out->m_Fill0Style);
	}

	if (out->m_StateFillStyle1)
	{
		uncompData += ReadBitData(uncompData, out->m_Fill1Style, numFillBits, bitOffset);
		dprintf("\t Fill1Style: %d\n", out->m_Fill1Style);
	}

	if (out->m_StateLineStyle)
	{
		uncompData += ReadBitData(uncompData, out->m_LineStyle, numLineBits, bitOffset);
		dprintf("\t LineStyle: %d\n", out->m_LineStyle);
	}

	if (out->m_StateNewStyles)
	{
		uncompData += ReadFillStyleArray(uncompData, &out->m_FillStyleArray, shapeType);
		uncompData += ReadLineStyleArray(uncompData, &out->m_LineStyleArray, shapeType);

		unsigned char Bits = uncompData[0];
		int fillBits = Bits&0xF;
		int lineBits = (Bits>>4)&0xF;
		uncompData+=1;
	}

	int bytesRead = (int)(uncompData - startAddr);

	return bytesRead;
}


//------------------------------------------------------------------------------


int ReadShapeRecordType2(char* uncompData, stShapeRecordType2* out, int& bitOffset)
{
	int val = 0;
	int numBitsToRead = 0;
	float result = 0.0f;
	char* startAddr = uncompData;

	dprintf("\tShapeRecordType2:\n");

	uncompData += ReadBitData(uncompData, out->m_TypeFlag, 1, bitOffset);
	dprintf("\t TypeFlag: %d\n", out->m_TypeFlag);

	int EdgeFlag = 0;
	uncompData += ReadBitData(uncompData, EdgeFlag, 1, bitOffset);
	dprintf("\t EdgeFlag: %d\n", EdgeFlag);
	out->m_EdgeFlag = EdgeFlag;

	if (EdgeFlag==0) // EdgeRecordType0
	{
		EdgeRecordType0* edge0 = &out->m_EdgeRecord0;

		int numBits = 0;
		uncompData += ReadBitData(uncompData, numBits, 4, bitOffset);
		dprintf("\t NumBits: %d\n", numBits);
		edge0->m_NumBits = numBits;

		uncompData += ReadBitData(uncompData, edge0->m_ControlDeltaX, numBits, bitOffset);
		dprintf("\t ControlDeltaX: %d\n", edge0->m_ControlDeltaX);
		

		uncompData += ReadBitData(uncompData, edge0->m_ControlDeltaY, numBits, bitOffset);
		dprintf("\t ControlDeltaY: %d\n", edge0->m_ControlDeltaY);

		uncompData += ReadBitData(uncompData, edge0->m_AnchorDeltaX, numBits, bitOffset);
		dprintf("\t AnchorDeltaX: %d\n", edge0->m_AnchorDeltaX);

		uncompData += ReadBitData(uncompData, edge0->m_AnchorDeltaY, numBits, bitOffset);
		dprintf("\t AnchorDeltaY: %d\n", edge0->m_AnchorDeltaY);
	}
	else // EdgeRecordType1
	{
		EdgeRecordType1* edge1 = &out->m_EdgeRecord1;

		int numBits = 0;
		uncompData += ReadBitData(uncompData, numBits, 4, bitOffset);
		dprintf("\t NumBits: %d\n", numBits);
		edge1->m_NumBits = numBits;

		int LineFlag = 0;
		uncompData += ReadBitData(uncompData, LineFlag, 1, bitOffset);
		dprintf("\t LineFlag: %d\n", LineFlag);
		edge1->m_GeneralLineFlag = LineFlag;

		if (LineFlag)
		{
			uncompData += ReadBitData(uncompData, edge1->m_GLineDeltaX, numBits + 2, bitOffset);
			dprintf("\t GLineDeltaX: %d\n", edge1->m_GLineDeltaX);
			
			uncompData += ReadBitData(uncompData, edge1->m_GLineDeltaY, numBits + 2, bitOffset);
			dprintf("\t GLineDeltaY: %d\n", edge1->m_GLineDeltaY);
		}
		else
		{

			int VLineFlag = 0;
			uncompData += ReadBitData(uncompData, VLineFlag, 1, bitOffset);
			dprintf("\t VLineFlag: %d\n", VLineFlag);
			edge1->m_VertLineFlag = VLineFlag;

			if (VLineFlag==0)
			{
				uncompData += ReadBitData(uncompData, edge1->m_VLineDeltaX, numBits + 2, bitOffset);
				// We add 1, as its +2 bits bigger than numbits ;)
				bool neg = ((edge1->m_VLineDeltaX >> (numBits+1)) & 0x1);
				if (neg)
				{
					// add extra 1's to make it negative
					for (int i=numBits+1; i<32; i++)
					{
						edge1->m_VLineDeltaX |= (1<<i);
					}
					edge1->m_VLineDeltaX = (~edge1->m_VLineDeltaX)+1;
					edge1->m_VLineDeltaX = -edge1->m_VLineDeltaX;
				}
				dprintf("\t VLineDeltaX: %d\n", edge1->m_VLineDeltaX);
			}
			else
			{
				uncompData += ReadBitData(uncompData, edge1->m_VLineDeltaY, numBits + 2, bitOffset);
				// We add 1, as its +2 bits bigger than numbits ;)
				bool neg = ((edge1->m_VLineDeltaY >> (numBits+1)) & 0x1);
				if (neg)
				{
					// add extra 1's to make it negative
					for (int i=numBits+1; i<32; i++)
					{
						edge1->m_VLineDeltaY |= (1<<i);
					}
					edge1->m_VLineDeltaY = (~edge1->m_VLineDeltaY)+1;
					edge1->m_VLineDeltaY = -edge1->m_VLineDeltaY;
				}
				dprintf("\t VLineDeltaY: %d\n", edge1->m_VLineDeltaY);
			}
		}

	}

	int bytesRead = (int)(uncompData - startAddr);

	return bytesRead;
}




//------------------------------------------------------------------------------


int ReadShapeRecords(char * uncompData, stShapeWithStyle* out, enShapeType shapeType)
{
	int bitOffset = 0;

	while (true) // ShapeRecordType0 terminates with break
	{
		// Reading Shape Record
		//int typeFlag = (uncompData[0]>>8) & 0x1;
		char * tmp = uncompData;
		int  tmpOffset = bitOffset;
		int typeFlag = 0;
		tmp += ReadBitData(tmp, typeFlag, 1, tmpOffset);

		if (typeFlag==0)
		{
			// ShapeRecordType1 or ShapeRecordType0
			//int chosen = (uncompData[0]>>2) & 0x1F;
			int chosen = 0;
			int val = 0;
			for (int i=0; i<5; i++)
			{
				tmp += ReadBitData(tmp, val, 1, tmpOffset);
				chosen |= val;
			}

			if (chosen)
			{
				// ShapeRecordType1
				dprintf("\t ShapeRecordType1\n");
				stShapeRecordType1* sr = new stShapeRecordType1();
				out->AddToList(sr);
				uncompData += ReadShapeRecordType1(uncompData, out, sr, bitOffset, shapeType);
			}
			else
			{
				// ShapeRecordType0
				dprintf("\t ShapeRecordType0\n");
				dprintf("\t End Of Record\n");
				stShapeRecordType0* sr = new stShapeRecordType0();
				out->AddToList(sr);

				break;
			}
		}
		else
		{
			// ShapeRecordType2
			dprintf("\t ShapeRecordType2\n");
			stShapeRecordType2* sr = new stShapeRecordType2();
			out->AddToList(sr);
			uncompData += ReadShapeRecordType2(uncompData, sr, bitOffset);
		}
	}


	return 0;
}


//------------------------------------------------------------------------------

int ReadFillStyleArray(char * uncompData, stFillStyleArray* out, enShapeType shapeType)
{
	char* startAddr = uncompData;
	if (uncompData[0]==0xff)
	{
		out->m_FillStyleCount = *((unsigned short*)&uncompData[0]);
		uncompData+=2;
	}
	else
	{
		out->m_FillStyleCount = uncompData[0];
		uncompData+=1;
	}


	if (out->m_FillStyleCount>0)
	{
		out->m_FillStyles = new stFillStyle[out->m_FillStyleCount];
		for (int i=0; i<out->m_FillStyleCount; i++)
		{
			uncompData += ReadFillStyle(uncompData, &out->m_FillStyles[i], shapeType);
		}
	}

	int bytesRead = (int)(uncompData - startAddr);
	return bytesRead;
}

//------------------------------------------------------------------------------

int ReadLineStyleArray(char * uncompData, stLineStyleArray* out, enShapeType shapeType)
{
	char* startAddr = uncompData;
	if (uncompData[0]==0xff)
	{
		out->m_LineStyleCount = *((unsigned short*)&uncompData[0]);
		uncompData+=2;
	}
	else
	{
		out->m_LineStyleCount = uncompData[0];
		uncompData+=1;
	}

	if (out->m_LineStyleCount>0)
	{
		out->m_LineStyles = new stLineStyle[out->m_LineStyleCount];
		for (int i=0; i<out->m_LineStyleCount; i++)
		{
			uncompData += ReadLineStyle(uncompData, &out->m_LineStyles[i], shapeType);
		}
	}


	int bytesRead = (int)(uncompData - startAddr);
	return bytesRead;
}


//------------------------------------------------------------------------------

int ReadShapeWithStyle(char * uncompData, stShapeWithStyle* out, enShapeType shapeType)
{
	uncompData += ReadFillStyleArray(uncompData, &out->m_FillStyleArray, shapeType);
	uncompData += ReadLineStyleArray(uncompData, &out->m_LineStyleArray, shapeType);

	out->m_NumFillBits = (uncompData[0]>>4) & 0x0F;
	out->m_NumLineBits = (uncompData[0]) & 0x0F;
	uncompData+=1;

	dprintf("\tstShape: NumFillBits: %d,  NumLineBits: %d\n", 
													out->m_NumFillBits,
													out->m_NumLineBits);

	ReadShapeRecords(uncompData, out, shapeType);

	return 0;
};

//------------------------------------------------------------------------------

int ReadDefineShape(char * uncompData, stDefineShape* out, int tagLen)
{
	out->m_characterID = *((unsigned short*)&uncompData[0]);
	uncompData += 2;
	dprintf("\tstDefineShape : CharacterID: %d\n", out->m_characterID);

	dprintf("\t\t: ShapeBounds(Rect):\n");
	uncompData += ReadRect(uncompData, &out->m_shapeBounds);

	ReadShapeWithStyle(uncompData, &out->m_shapeWithStyle, SHAPE1);

	return 0;
}

//------------------------------------------------------------------------------

int ReadRemoveObject2(char * uncompData, stRemoveObject2* out, int tagLen)
{
	dprintf("\tRemoveObject2:\n");
	out->m_Depth = uncompData[0];

	return 0;
}

//------------------------------------------------------------------------------

int ReadPlaceObject2(char * uncompData, stPlaceObject2* out, int tagLen)
{
	dprintf("\tPlaceObject2:\n");

	int val = 0;
	int numBitsToRead = 0;
	int bitOffset = 0;
	float result = 0.0f;
	char* startAddr = uncompData;

	uncompData += ReadBitData(uncompData, val, 2, bitOffset);
	out->m_PlaceFlagsReserved = val;
	dprintf("\t PlaceFlagsReserved: %d\n", out->m_PlaceFlagsReserved);
	

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_PlaceFlagHasName = val;
	dprintf("\t PlaceFlagHasName: %d\n", out->m_PlaceFlagHasName);
	

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_PlaceFlagHasRatio = val;
	dprintf("\t PlaceFlagHasRatio: %d\n", out->m_PlaceFlagHasRatio);
	

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_PlaceFlagHasColorTransform = val;
	dprintf("\t PlaceFlagHasColorTransform: %d\n", out->m_PlaceFlagHasColorTransform);
	

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_PlaceFlagHasMatrix = val;
	dprintf("\t PlaceFlagHasMatrix: %d\n", out->m_PlaceFlagHasMatrix);

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_PlaceFlagHasCharacter = val;
	dprintf("\t PlaceFlagHasCharacter: %d\n", out->m_PlaceFlagHasCharacter);
	

	uncompData += ReadBitData(uncompData, val, 1, bitOffset);
	out->m_PlaceFlagMove = val;
	dprintf("\t PlaceFlagMove: %d\n", out->m_PlaceFlagMove);
	

	//uncompData += ReadBitData(uncompData, val, 16, bitOffset);
	// ISSUE - byte little big endian with values of 16 etc
	unsigned short Depth = ((unsigned short*)uncompData)[0];
	uncompData+=2;
	out->m_Depth = Depth;
	dprintf("\t Depth: %d\n", Depth);

	if (out->m_PlaceFlagHasCharacter)
	{
		out->m_CharacterId = ((unsigned short*)uncompData)[0];
		dprintf("\t CharacterID: %d\n", out->m_CharacterId);
		uncompData+=2;
	}

	if (out->m_PlaceFlagHasMatrix)
	{
		stMatrix* mat = &out->m_Matrix;
		uncompData += ReadMatrix(uncompData, mat);
	}

	if (out->m_PlaceFlagHasColorTransform)
	{
		stCXForm* ct = &out->m_ColorTransform;
		uncompData += ReadCXForm(uncompData, ct);
		int jjj = 2;
	}

	if (out->m_PlaceFlagHasRatio)
	{
		out->m_Ratio = ((unsigned short*)uncompData)[0];
		uncompData+=2;
		dprintf("\t Ratio: %d\n", out->m_Ratio);
	}

	if (out->m_PlaceFlagHasName)
	{
		memset(out->m_Name, 0, 32);
		strcpy(out->m_Name, uncompData);
		dprintf("\t name: %s\n", out->m_Name);
	}

	return 0;
}


//------------------------------------------------------------------------------

int ReadDefineSprite(char * uncompData, stDefineSprite* out, int tagLen)
{
	out->m_characterID = *((unsigned short*)&uncompData[0]);
	uncompData += 2;

	out->m_FrameCount = *((unsigned short*)&uncompData[0]);
	uncompData += 2;

	dprintf("\tDefineSprite:\n");
	dprintf("\t characterID: %d\n", out->m_characterID);
	dprintf("\t FrameCount: %d\n",  out->m_FrameCount);

	char * tmpData = uncompData;
	int    tmpLen  = tagLen-4;

	ReadDataChunks(tmpData, tmpLen, &out->m_tags);

	return 0;
}


//------------------------------------------------------------------------------


void ReadDataChunks(char * data, int lenData, stSWFTags** parentTag)
{
	char * tmpData = data;
	int    tmpLen  = lenData;

	while (tmpLen > 0)
	{
		stSWFTags* tag = AddTag(parentTag);

		unsigned short int tmpx = tmpData[1] | (tmpData[0]<<8);
		unsigned short int *tmp0 = (unsigned short int*)&tmpData[0];
		tmpData+=2;
		tmpLen-=2;

		int Tag = (*tmp0 >> 6);
		int Length = (*tmp0 & 0x3F);

		if (Length != 0x3F)
		{
			// Small Tag Data
			dprintf("SmallTag ID: 0x%X, Length:%d\n", Tag, Length);

			tag->m_id = Tag;
			tag->m_length = Length;
			tag->m_tagType = stSWFTags::tagSmall;
		}
		else
		{
			// Large Tag Data
			Length =  *((unsigned int*)&tmpData[0]);
			tmpData+=4;
			tmpLen-=4;

			dprintf("LargeTag ID: 0x%X, Length:%d\n", Tag, Length);
			tag->m_id = Tag;
			tag->m_length = Length;
			tag->m_tagType = stSWFTags::tagLarge;
		}

		char * tagName = "Unknown";
		for (int i=0; i<sizeof(Tags)/sizeof(stTagType); i++)
		{
			if (Tag == Tags[i].m_id)
			{
				tagName = Tags[i].m_name;
				break;
			}
		}

		strcpy( tag->m_name, tagName );

		dprintf(" +- Name %s, RawBytes:0x%.2X, 0x%.2X\n",
													tagName,
													data[0],
													data[1]);


		ReadDataTag(Tag, Length, tmpData, tag);

		tmpData+=Length;
		tmpLen-=Length;
	}

	dprintf("\nEnd Reading Tags\n");
}


//------------------------------------------------------------------------------

void ReadDataTag(int tagID, int tagLength, char * data, stSWFTags * tag)
{
	switch (tagID)
	{
		case TagID_SETBACKGROUNDCOLOR:
		{
			stBackGroundColour*  backGroundColour = new stBackGroundColour();
			tag->m_base = backGroundColour;
			ReadBackGroundColour(data, backGroundColour);
		}
		break;

		case TagID_FRAMELABEL:
		{
			stFrameLabel* frameLabel = new stFrameLabel();
			tag->m_base = frameLabel;
			ReadFrameLabel(data, frameLabel);
		}
		break;

		case TagID_DEFINEBITSJPEG2:
		{
			stDefineBitsJPEG2* jpg = new stDefineBitsJPEG2();
			tag->m_base = jpg;
			ReadDefineBitsJPEG2(data, jpg, tagLength);
		}
		break;

		case TagID_DEFINEBITSJPEG3:
		{
			stDefineBitsJPEG3*  jpg = new stDefineBitsJPEG3();
			tag->m_base = jpg;
			ReadDefineBitsJPEG3(data, jpg, tagLength);
		}
		break;

		case TagID_DEFINESHAPE:
		{
			stDefineShape* shape = new stDefineShape();
			tag->m_base = shape;
			ReadDefineShape(data, shape, tagLength);
		}
		break;

		case TagID_DEFINESPRITE:
		{
			stDefineSprite* sprite = new stDefineSprite();
			tag->m_base = sprite;
			ReadDefineSprite(data, sprite, tagLength);
		}
		break;

		// -- Sprite Tag -----
		case TagID_PLACEOBJECT2:
		{
			stPlaceObject2* po = new stPlaceObject2();
			tag->m_base = po;
			ReadPlaceObject2(data, po, tagLength);
		}
		break;

		case TagID_SHOWFRAME:
		{
			dprintf("\tShowFrame\n");
		}
		break;

		case TagID_END:
		{
			dprintf("\tEnd\n");
		}
		break;

		case TagID_DEFINEFONT2:
		{
			dprintf("\tSkip DefineFont2\n");
		}
		break;

		case TagID_DEFINEEDITTEXT:
		{
			dprintf("\tSkip DefineEditText\n");
		}
		break;

		case TagID_REMOVEOBJECT2:
		{
			stRemoveObject2* ro = new stRemoveObject2();
			tag->m_base = ro;
			ReadRemoveObject2(data, ro, tagLength);
		}
		break;

		default:
		{
			// unknown
			dprintf("unknown\n");
		}
	}
}


//------------------------------------------------------------------------------

int ReadFlash(char * szFileName, stSWF* swf)
{
	dprintf("Reading Flash File :%s\n", szFileName);


	FILE * fp = fopen(szFileName, "rb");

	extern int FileSize(FILE *fp); // loadjpg.cpp
	int inSize  = FileSize(fp);

	stSWFHeader* head = &swf->m_header;

	dprintf("---\n");

	fread (	&head->m_Signature,			// void * buffer,
			3 * sizeof(char),			// size
			1,							// count,
			fp );						// FILE * stream

	fread (	&head->m_Version,			// void * buffer,
			sizeof(head->m_Version),	// size
			1,							// count,
			fp );						// FILE * stream

	fread (	&head->m_FileLength,		// void * buffer,
			sizeof(head->m_FileLength),	// size
			1,							// count,
			fp );						// FILE * stream

	unsigned char * sig = head->m_Signature;

	dprintf("stSWFHeader:\n");
	dprintf(" +- Signature: %c %c %c\n",			sig[0], sig[1], sig[2]);
	dprintf(" +- Version: %d\n",					head->m_Version);
	dprintf(" +- Uncompressed FileLength: %d\n",	head->m_FileLength);


	// If its compressed! - which we know it is for this example :)
	dprintf(" o- Decompressing...\n");

	int (*uncompress) (char *dest, unsigned int *destLen, const char *source, unsigned int sourceLen);
	HMODULE dll = LoadLibrary("zlib1.dll");
	uncompress = (int (*) (char *dest, unsigned int *destLen, const char *source, unsigned int sourceLen))GetProcAddress(dll, "uncompress");


	

	// Just allocate a big junk of memory, read it all in, and uncompress it!
	int outSize = head->m_FileLength + 50;
	char * tmpIn  = new char [inSize + 50]; // +50 is just some padding ;)
	char * tmpOut = new char [outSize];


	memset(tmpIn,  0, inSize);
	memset(tmpOut, 0, outSize);


	// Read the whole file data in
	int done = 1;
	int size = 0;
	while (done)
	{
		done = (int)fread(tmpIn + size, 1, 1, fp);
		if (done)
		{
			size += 1;
		}
	}

	unsigned int tmpOutLen = outSize;
	char * uncompData = NULL;

	if (sig[0]=='C' &&
		sig[1]=='W' &&
		sig[2]=='S' )
	{
		int error = uncompress(tmpOut, &tmpOutLen, tmpIn, size);

		delete[] tmpIn;
		tmpIn = NULL;

		uncompData = tmpOut;
	}
	else if (sig[0]=='F' &&
			 sig[1]=='W' &&
			 sig[2]=='S' )
	{
		delete[] tmpOut;
		tmpOut=NULL;

		tmpOutLen  = size;
		uncompData = tmpIn;
	}
	else
	{
		dprintf("ERROR-UNKNOWN FORMAT\n");
		return 1;
	}

	//-----------------------------

	dprintf(" o- Decompressing Done\n");

	//-----------------------------

	dprintf(" +- FrameSize (Rect)\n");
	stRect rect;
	int numBytes = ReadRect(uncompData, &rect);
	uncompData += numBytes;
	head->m_FrameSize = rect;

	unsigned short int *tmpInt = (unsigned short int*)&uncompData[0];
	head->m_FrameRate = *tmpInt;
	uncompData+=2;
	tmpOutLen-=2;

	tmpInt = (unsigned short int*)&uncompData[0];
	head->m_FrameCount = *tmpInt;
	uncompData+=2;
	tmpOutLen-=2;

	dprintf(" +- Head FrameRate: %d\n",		head->m_FrameRate);
	dprintf(" +- Head FrameCount: %d\n\n",	head->m_FrameCount);

	//-----------------------------

	dprintf("*Reading Chunks*\n");

	ReadDataChunks(&uncompData[0], tmpOutLen, &swf->m_tags);

	//-----------------------------

	if (tmpOut)
	{
		delete[] tmpOut;
		tmpOut = NULL;
	}
	if (tmpIn)
	{
		delete[] tmpIn;
		tmpIn = NULL;
	}

	fclose(fp);

	dprintf("done\n");
	return 0;
}

//------------------------------------------------------------------------------


