/******************************************************************************/
/*                                                                            */
/*  File: swf.h                                                               */
/*  Author: bkenwright@xbdev.net                                              */
/*  Date: 19-01-06                                                            */
/*                                                                            */
/******************************************************************************/
/*
    swf file format loader/parser

	The flash swf file format is a great format for animations, and simple
	2d.  This code is a simple example of how you can parse the flash
	swf file format.  It dumps all the data to a txt file so you can
	examine it, and also loads it into a set of data containers where you
	can use it later to render or display.
*/
/******************************************************************************/

#ifndef SWF_H
#define SWF_H

#include <stdio.h>
#include <stdarg.h>

//------------------------------------------------------------------------------

enum enTagIDs
{
	TagID_END                =  0,   
	TagID_SHOWFRAME          =  1,   
	TagID_DEFINESHAPE        =  2,   
	TagID_FREECHARACTER      =  3,   
	TagID_PLACEOBJECT        =  4,   
	TagID_REMOVEOBJECT       =  5,   
	TagID_DEFINEBITS         =  6,   
	TagID_DEFINEBUTTON       =  7,   
	TagID_JPEGTABLES         =  8,   
	TagID_SETBACKGROUNDCOLOR =  9,   
	TagID_DEFINEFONT         =  10,  
	TagID_DEFINETEXT         =  11,  
	TagID_DOACTION           =  12,  
	TagID_DEFINEFONTINFO     =  13,  
	TagID_DEFINESOUND        =  14,  
	TagID_STARTSOUND         =  15,  
	TagID_DEFINEBUTTONSOUND  =  17,  
	TagID_SOUNDSTREAMHEAD    =  18,  
	TagID_SOUNDSTREAMBLOCK   =  19,  
	TagID_DEFINELOSSLESS     =  20,  
	TagID_DEFINEBITSJPEG2    =  21,  
	TagID_DEFINESHAPE2       =  22,  
	TagID_DEFINEBUTTONCXFORM =  23,  
	TagID_PROTECT            =  24,  
	TagID_PATHSAREPOSTSCRIPT =  25,  
	TagID_PLACEOBJECT2       =  26,  
	TagID_REMOVEOBJECT2      =  28,  
	TagID_SYNCFRAME          =  29,  
	TagID_FREEALL            =  31,  
	TagID_DEFINESHAPE3       =  32,  
	TagID_DEFINETEXT2        =  33,  
	TagID_DEFINEBUTTON2      =  34,  
	TagID_DEFINEBITSJPEG3    =  35,  
	TagID_DEFINELOSSLESS2    =  36,  
	TagID_DEFINEEDITTEXT     =  37,  
	TagID_DEFINEVIDEO        =  38,  
	TagID_DEFINESPRITE       =  39,  
	TagID_NAMECHARACTER      =  40,  
	TagID_SERIALNUMBER       =  41,  
	TagID_DEFINETEXTFORMAT   =  42,  
	TagID_FRAMELABEL         =  43,  
	TagID_SOUNDSTREAMHEAD2   =  45,  
	TagID_DEFINEMORPHSHAPE   =  46,  
	TagID_FRAMETAG           =  47,  
	TagID_DEFINEFONT2        =  48,  
	TagID_GENCOMMAND         =  49,  
	TagID_DEFINECOMMANDOBJ   =  50,  
	TagID_CHARACTERSET       =  51,  
	TagID_FONTREF            =  52,  
	TagID_EXPORTASSETS       =  56,  
	TagID_IMPORTASSETS       =  57,  
	TagID_ENABLEDEBUGGER     =  58,  
	TagID_INITACTION         =  59,  
	TagID_DEFINEVIDEOSTREAM  =  60,  
	TagID_VIDEOFRAME         =  61,  
	TagID_DEFINEFONTINFO2    =  62,  
	TagID_DEFINEBITSPTR      =  1023,

};

struct stTagType
{
	int		m_id;
	char	m_name[32];
};

static stTagType Tags [] =
{
	{ TagID_END                ,  "END"					},                    
	{ TagID_SHOWFRAME          ,  "SHOWFRAME"			},              
	{ TagID_DEFINESHAPE        ,  "DEFINESHAPE"			},            
	{ TagID_FREECHARACTER      ,  "FREECHARACTER"		},          
	{ TagID_PLACEOBJECT        ,  "PLACEOBJECT"			},            
	{ TagID_REMOVEOBJECT       ,  "REMOVEOBJECT"		},           
	{ TagID_DEFINEBITS         ,  "DEFINEBITS"			},             
	{ TagID_DEFINEBUTTON       ,  "DEFINEBUTTON"		},           
	{ TagID_JPEGTABLES         ,  "JPEGTABLES"			},             
	{ TagID_SETBACKGROUNDCOLOR ,  "SETBACKGROUNDCOLOR"	},     
	{ TagID_DEFINEFONT         ,  "DEFINEFONT"			},             
	{ TagID_DEFINETEXT         ,  "DEFINETEXT"			},             
	{ TagID_DOACTION           ,  "DOACTION"			},              
	{ TagID_DEFINEFONTINFO     ,  "DEFINEFONTINFO"		},        
	{ TagID_DEFINESOUND        ,  "DEFINESOUND"			},           
	{ TagID_STARTSOUND         ,  "STARTSOUND"			},            
	{ TagID_DEFINEBUTTONSOUND  ,  "DEFINEBUTTONSOUND"	},     
	{ TagID_SOUNDSTREAMHEAD    ,  "SOUNDSTREAMHEAD"		},       
	{ TagID_SOUNDSTREAMBLOCK   ,  "SOUNDSTREAMBLOCK"	},      
	{ TagID_DEFINELOSSLESS     ,  "DEFINELOSSLESS"		},        
	{ TagID_DEFINEBITSJPEG2    ,  "DEFINEBITSJPEG2"		},       
	{ TagID_DEFINESHAPE2       ,  "DEFINESHAPE2"		},          
	{ TagID_DEFINEBUTTONCXFORM ,  "DEFINEBUTTONCXFORM"	},     
	{ TagID_PROTECT            ,  "PROTECT"				},               
	{ TagID_PATHSAREPOSTSCRIPT ,  "PATHSAREPOSTSCRIPT" },     
	{ TagID_PLACEOBJECT2       ,  "PLACEOBJECT2"		},          
	{ TagID_REMOVEOBJECT2      ,  "REMOVEOBJECT2"		},         
	{ TagID_SYNCFRAME          ,  "SYNCFRAME"			},             
	{ TagID_FREEALL            ,  "FREEALL"				},               
	{ TagID_DEFINESHAPE3       ,  "DEFINESHAPE3"		},          
	{ TagID_DEFINETEXT2        ,  "DEFINETEXT2"			},           
	{ TagID_DEFINEBUTTON2      ,  "DEFINEBUTTON2"		},         
	{ TagID_DEFINEBITSJPEG3    ,  "DEFINEBITSJPEG3"		},       
	{ TagID_DEFINELOSSLESS2    ,  "DEFINELOSSLESS2"		},       
	{ TagID_DEFINEEDITTEXT     ,  "DEFINEEDITTEXT"		},        
	{ TagID_DEFINEVIDEO        ,  "DEFINEVIDEO"			},           
	{ TagID_DEFINESPRITE       ,  "DEFINESPRITE"		},          
	{ TagID_NAMECHARACTER      ,  "NAMECHARACTER"		},         
	{ TagID_SERIALNUMBER       ,  "SERIALNUMBER"		},          
	{ TagID_DEFINETEXTFORMAT   ,  "DEFINETEXTFORMAT"	},      
	{ TagID_FRAMELABEL         ,  "FRAMELABEL"			},            
	{ TagID_SOUNDSTREAMHEAD2   ,  "SOUNDSTREAMHEAD2"	},      
	{ TagID_DEFINEMORPHSHAPE   ,  "DEFINEMORPHSHAPE"	},      
	{ TagID_FRAMETAG           ,  "FRAMETAG"			},              
	{ TagID_DEFINEFONT2        ,  "DEFINEFONT2"			},           
	{ TagID_GENCOMMAND         ,  "GENCOMMAND"			},            
	{ TagID_DEFINECOMMANDOBJ   ,  "DEFINECOMMANDOBJ"	},      
	{ TagID_CHARACTERSET       ,  "CHARACTERSET"		},          
	{ TagID_FONTREF            ,  "FONTREF"				},               
	{ TagID_EXPORTASSETS       ,  "EXPORTASSETS"		},          
	{ TagID_IMPORTASSETS       ,  "IMPORTASSETS"		},          
	{ TagID_ENABLEDEBUGGER     ,  "ENABLEDEBUGGER"		},        
	{ TagID_INITACTION         ,  "INITACTION"			},            
	{ TagID_DEFINEVIDEOSTREAM  ,  "DEFINEVIDEOSTREAM"	},     
	{ TagID_VIDEOFRAME         ,  "VIDEOFRAME"			},            
	{ TagID_DEFINEFONTINFO2    ,  "DEFINEFONTINFO2"		},        
	{ TagID_DEFINEBITSPTR      ,  "DEFINEBITSPTR"		},         
};

//------------------------------------------------------------------------------

#pragma pack(1)
struct stSmallTag
{
	unsigned short m_Tag    : 10;		// UB[10] Tag id 
	unsigned short m_Length : 6;		// UB[6] Length of tag 
};
#pragma pack()

//------------------------------------------------------------------------------

#pragma pack(1)
struct stLargeTag
{
	unsigned short int m_Tag	  : 10;	// UB[10] Tag id 
	unsigned short int m_LengthID : 6;	// Long Header Flag UB[6] Always 0x3F  
	int				   m_Length;		// UI32 Length of tag 
};
#pragma pack()

//------------------------------------------------------------------------------

#pragma pack(1)
struct stRect
{
	int m_Nbits;						// nBits = UB[5] Bits in each rect value field 
	union
	{
		int m_sizes[4];
		struct  
		{
			int	m_Xmin;					// SB[nBits] X minimum position for rect 
			int m_Xmax;					// SB[nBits] X maximum position for rect 
			int m_Ymin;					// SB[nBits] Y minimum position for rect 
			int m_Ymax;					// SB[nBits] Y maximum position for rect 
		};
	};
};
#pragma pack()

//------------------------------------------------------------------------------

#pragma pack(1)
struct stSWFHeader
{
	unsigned char		m_Signature[3];	// UI8 Signature byte 1 - always 'F' 'W' 'S' 
	unsigned char		m_Version;		// UI8 Single byte file version 
	int					m_FileLength;	// UI32 Length of entire file in bytes (uncompressed size)

	// -- Uncompress from here --

	stRect				m_FrameSize;	// RECT Frame size in TWIPS 
	unsigned short int	m_FrameRate;	// UI16 Frame delay in 8.8 fixed number of frames per second 
	unsigned short int	m_FrameCount;	// UI16 Total number of frames in movie 
};
#pragma pack()


//------------------------------------------------------------------------------

struct stSWFBase
{
	stSWFBase(int id){m_ID=id;};
	virtual ~stSWFBase(){};
	int m_ID;
};

//------------------------------------------------------------------------------

struct stSWFTags
{
	enum tagType
	{
		tagSmall,
		tagLarge
	};

	int			m_id;
	int			m_length;
	int			m_tagType;
	char		m_name[32];
	stSWFBase*	m_base;

	stSWFTags*  m_next;

	stSWFTags()
	{
		m_id		= -1;
		m_length	= 0;
		m_tagType	= -1;
		m_base		= NULL;
		m_next		= NULL;
	};
	~stSWFTags()
	{
		if (m_base)
		{
			delete m_base;
			m_base=NULL;
		}
	}
};


//------------------------------------------------------------------------------

struct stMatrix
{
	int		m_hasScale;			// 1
	int		m_numScaleBits;		// 5
	float	m_xScale;			// ~
	float	m_yScale;			// ~
	int		m_hasRotate;		// 1
	int		m_numRotateBits;	// 5
	float	m_RotateSkew0;		// ~
	float	m_RotateSkew1;		// ~
	int		m_numTranslateBits; // 5
	int		m_xTranslate;		// ~
	int		m_yTranslate;		// ~
};

struct stCXForm
{
	int		m_HasMultTerms;		// hasMult = UB[1] Has color multiply values if equal to 1 
	int		m_HasAddTerms;		// hasAdd = UB[1] Has color addition values if equal to 1 
	int		m_Nbits;			// nBits = UB[4] Bits in each value field 
	float	m_RedMultTerm;		// If hasMult SB[nBits] Red multiply value 
	float	m_GreenMultTerm;	// If hasMult SB[nBits] Green multiply value 
	float	m_BlueMultTerm;		// If hasMult SB[nBits] Blue multiply value 
	float   m_AlphaMultTerm;
	int		m_RedAddTerm;		// If hasAdd SB[nBits] Red addition value 
	int		m_GreenAddTerm;		// If hasAdd SB[nBits] Green addition value 
	int		m_BlueAddTerm;		// If hasAdd SB[nBits] Blue addition value 
	int		m_AlphaAddTerm;		// If hasAdd SB[nBits] Blue addition value 

	stCXForm()
	{
		m_AlphaAddTerm=m_BlueAddTerm=m_GreenAddTerm=m_RedAddTerm=0;
		m_AlphaMultTerm=m_BlueMultTerm=m_GreenMultTerm=m_RedMultTerm=1.0f;
		m_HasMultTerms=m_HasAddTerms=0;
		m_Nbits = 0;
	}

	const stCXForm operator* (stCXForm& f)
	{
		stCXForm theColour;
		
		theColour.m_AlphaMultTerm = m_AlphaMultTerm * f.m_AlphaMultTerm;
		theColour.m_RedMultTerm   = m_RedMultTerm   * f.m_RedMultTerm;
		theColour.m_GreenMultTerm = m_GreenMultTerm * f.m_GreenMultTerm;
		theColour.m_BlueMultTerm  = m_BlueMultTerm  * f.m_BlueMultTerm;
		theColour.m_HasMultTerms  = m_HasMultTerms | f.m_HasMultTerms;

		if ( m_RedAddTerm != 0 )
		{
			int jjj = 2;
		}

		if ( f.m_RedAddTerm != 0)
		{
			int jjs=2;
		}

		theColour.m_AlphaAddTerm = (int)(m_AlphaAddTerm * f.m_AlphaMultTerm) + f.m_AlphaAddTerm;
		theColour.m_RedAddTerm   = (int)(m_RedAddTerm   * f.m_RedMultTerm)   + f.m_RedAddTerm;
		theColour.m_GreenAddTerm = (int)(m_GreenAddTerm * f.m_GreenMultTerm) + f.m_GreenAddTerm;
		theColour.m_BlueAddTerm  = (int)(m_BlueAddTerm  * f.m_BlueMultTerm)  + f.m_BlueAddTerm;

		theColour.m_HasAddTerms  = m_HasAddTerms | f.m_HasAddTerms;

		//*this = theColour;

		return theColour;
		//return *this;
	}

	stCXForm operator= (const stCXForm& f)
	{
		if ( m_RedAddTerm != 0 )
		{
			int jjj = 2;
		}

		if ( f.m_RedAddTerm != 0)
		{
			int jjs=2;
		}

		if (this!=&f)
		{
			m_AlphaMultTerm = f.m_AlphaMultTerm;
			m_RedMultTerm   = f.m_RedMultTerm;
			m_GreenMultTerm = f.m_GreenMultTerm;
			m_BlueMultTerm  = f.m_BlueMultTerm;

			m_AlphaAddTerm = f.m_AlphaAddTerm;
			m_RedAddTerm   = f.m_RedAddTerm;
			m_GreenAddTerm = f.m_GreenAddTerm;
			m_BlueAddTerm  = f.m_BlueAddTerm;

			m_HasAddTerms  = f.m_HasAddTerms;
			m_HasMultTerms = f.m_HasMultTerms;

		}
		return *this;
	}
};
/*
const Matrix operator*(Matrix& m) { 
	assert(numCols==m.numRows); 
	Matrix theMatrix(numRows,m.numCols); 
	for (int r=0;r<numRows;r++) { 
		for (int c=0;c<m.numCols;c++) { 
			for (int i=0;i<numCols;i++) { 
				theMatrix[r][c]+=matrix[r][i]*m[i][c]; 
			} 
		} 
	} 
	return theMatrix; 
} 

Person& Person::operator=(const Person& p) {
	if (this != &p) {  // make sure not same object
		delete [] _name;                     // Delete old name's memory.
		_name = new char[strlen(p._name)+1]; // Get new space
		strcpy(_name, p._name);              // Copy new name
		_id = p._id;                         // Copy id
	}
	return *this;    // Return ref for multiple assignment
}//end operator=
*/

//------------------------------------------------------------------------------
/*
struct stEndShapeRecord
{
};

//------------------------------------------------------------------------------

struct stStyleChangeRecord
{
};

//------------------------------------------------------------------------------

struct stStraightEdgeRecord
{
};

//------------------------------------------------------------------------------

struct stCurvedEdgeRecord
{
};
*/
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------

struct EdgeRecordType0
{

	int m_NumBits;			// nBits = UB[4] + 2 Number of bits per value 
	int m_ControlDeltaX;	// SB[nBits] X control point change 
	int m_ControlDeltaY;	// SB[nBits] Y control point change 
	int m_AnchorDeltaX;		// SB[nBits] X anchor point change 
	int m_AnchorDeltaY;		// SB[nBits] Y anchor point change
};

//------------------------------------------------------------------------------

struct EdgeRecordType1
{
	int m_NumBits;			// nBits = UB[4] + 2 Number of bits per value 
	int m_GeneralLineFlag;	// lineFlag = UB[1] General Line equals 1 
	int m_GLineDeltaX;		// If lineFlag = 1 SB[nBits] X delta 
	int m_GLineDeltaY;		// If lineFlag = 1 SB[nBits] Y delta 
	int m_VertLineFlag;		// If lineFlag = 0 vertFlag = SB[1] Vertical Line equals 1 
	int m_VLineDeltaX;		// If vertFlag = 0 SB[nBits] X delta 
	int m_VLineDeltaY;		// If vertFlag = 1 SB[nBits] Y delta 
};

//------------------------------------------------------------------------------



struct stGradientRecord
{
	unsigned char m_Ratio;	// UI8 Ratio value 
	unsigned int  m_Colour;	// RGB (Shape1 or Shape2)
	// RGBA (Shape3) Color of gradient 
};

struct stGradient
{
	stGradient()
	{
		m_NumGradients=0;
		m_GradientRecords=NULL;
	}

	~stGradient()
	{
		if (m_GradientRecords)
		{
			delete[] m_GradientRecords;
			m_GradientRecords=NULL;
		}
	}

	int					m_NumGradients;			// 8
	stGradientRecord*	m_GradientRecords;
};



enum enShapeType
{
	SHAPE1 = 0,
	SHAPE2 = 1,
	SHAPE3 = 3
};

//------------------------------------------------------------------------------

struct stLineStyle
{
	unsigned short	m_Width;	// 16
	unsigned int	m_RGBA;		// 24 ~ 32
};

//------------------------------------------------------------------------------

enum enFillStyleType
{
	FS_SOLID	= 0x00,	// solid fill
	FS_LINEAR	= 0x10,	// linear gradient fill
	FS_RAIAL	= 0x12,	// radial gradient fill
	FS_TILED	= 0x40,	// tiled bitmap fill
	FS_CLIPPED	= 0x41,	// clipped bitmap fill
};

struct stFillStyle
{
	enFillStyleType m_FillStyleType;	// 8
	unsigned int	m_RGBA;				// 24 or 32
	stMatrix		m_GradientMatrix;	//
	stGradient		m_Gradient;
	unsigned int	m_BitmapID;			// 16
	stMatrix		m_BitmapMatrix;
};

//------------------------------------------------------------------------------


struct stFillStyleArray
{
	stFillStyleArray()
	{
		m_FillStyleCount=0;
		m_FillStyles=NULL;
	}
	~stFillStyleArray()
	{
		if (m_FillStyles)
		{
			delete[] m_FillStyles;
			m_FillStyles=NULL;
		}
	}
	int				m_FillStyleCount;
	stFillStyle*	m_FillStyles;
};

//------------------------------------------------------------------------------

struct stLineStyleArray
{
	stLineStyleArray()
	{
		m_LineStyleCount=0;
		m_LineStyles=NULL;
	}
	~stLineStyleArray()
	{
		if (m_LineStyles)
		{
			delete[] m_LineStyles;
			m_LineStyles=NULL;
		}
	}

	int				m_LineStyleCount;
	stLineStyle*	m_LineStyles;
};


//------------------------------------------------------------------------------

enum enShapeRecordType
{
	ShapeRecordEnd,
	ShapeRecord1,
	ShapeRecord2

};

struct stShapeRecordBase
{
	stShapeRecordBase(enShapeRecordType type)
	{
		m_shapeRecordType=type;
		m_next=NULL;
	}
	virtual ~stShapeRecordBase(){}

	int	m_shapeRecordType;
	stShapeRecordBase* m_next;
};

//------------------------------------------------------------------------------


struct stShapeRecordType0 : public stShapeRecordBase // End of shape record
{
	stShapeRecordType0() : stShapeRecordBase(ShapeRecordEnd){};

	int m_TypeFlag;			// UB[1] = 0 Non-edge record flag 
	int m_EndOfShape;		// UB[5] = 0 End of shape flag 
};

//------------------------------------------------------------------------------

struct stShapeRecordType1 : public stShapeRecordBase // Line Style, Fill Style and Positions
{
	stShapeRecordType1() : stShapeRecordBase(ShapeRecord1){};

	int m_TypeFlag;			// UB[1] = 0 Non-edge record flag 
	int m_StateNewStyles;	// newStyles = UB[1] New styles flag. Used by DefineShape2 and DefineShape3 only. 
	int m_StateLineStyle;	// lineStyle = UB[1] Line style change flag 
	int m_StateFillStyle1;	// fillStyle1 = UB[1] Fill style 1 change flag
	int m_StateFillStyle0;	// fillStyle0 = UB[1] Fill style 0 change flag

	int m_StateMoveTo;		// moveTo = UB[1] Move to flag 
	int m_MoveBits;			// If moveTo nMoveBits = UB[5] Move bit count 
	int m_MoveX;			// If moveTo SB[nMoveBits] X value (relative to shape origin)
	int m_MoveY;			// If moveTo SB[nMoveBits] Y value (relative to shape origin)

	int m_Fill0Style;		// If fillStyle0 UB[nFillBits] Fill 0 Style 
	int m_Fill1Style;		// If fillStyle1 UB[nFillBits] Fill 1 Style 
	int m_LineStyle;		// If lineStyle UB[nLineBits] Line Style 

	stFillStyleArray m_FillStyleArray; // If newStyles FILLSTYLEARRAY Array of new fill styles 
	stLineStyleArray m_LineStyleArray; // If newStyles LINESTYLEARRAY Array of new line styles

	int m_NumFillBits;		// If newStyles UB[4] Number of fill index bits for new styles
	int m_NumLineBits;		// If newStyles UB[4] Number of line index bits for new styles
};

//------------------------------------------------------------------------------

struct stShapeRecordType2 : public stShapeRecordBase // Straight or curved edge 
{
	stShapeRecordType2() : stShapeRecordBase(ShapeRecord2){};

	int m_TypeFlag;			// UB[1] = 1 This is an edge record 
	int m_EdgeFlag;			// UB[1] = 0 Curved edge - always 0 
							// UB[1] = 1 Straight edge - always 1 
	union
	{
		EdgeRecordType0 m_EdgeRecord0;
		EdgeRecordType1 m_EdgeRecord1;
	};
};

//------------------------------------------------------------------------------

struct stShapeWithStyle
{
	stShapeWithStyle()
	{
		m_ShapeRecord=NULL;
		m_NumFillBits=0;
		m_NumLineBits=0;
	}
	~stShapeWithStyle()
	{
		stShapeRecordBase* tmp = m_ShapeRecord;
		while (tmp)
		{
			stShapeRecordBase* nex = tmp->m_next;
			delete tmp;
			tmp=NULL;

			tmp=nex;
		}
	}

	void AddToList(stShapeRecordBase* addItem)
	{
		if (m_ShapeRecord==NULL)
		{
			m_ShapeRecord = addItem;
		}
		else
		{
			stShapeRecordBase* tmp = m_ShapeRecord;
			while ( tmp->m_next )
			{
				tmp = tmp->m_next;
			}
			tmp->m_next = addItem;
		}
	}

	stFillStyleArray m_FillStyleArray;
	stLineStyleArray m_LineStyleArray;

	int				m_NumFillBits;	// nFillBits = UB[4] Number of fill index bits 
	int				m_NumLineBits;	// nLineBits = UB[4] Number of line index bits 
	
	stShapeRecordBase*	m_ShapeRecord;	// SHAPEREC[one or more] Shape records 
};

//------------------------------------------------------------------------------


struct stDefineShape : public stSWFBase
{
	stDefineShape() : stSWFBase(TagID_DEFINESHAPE2){};

	unsigned short		m_characterID;
	stRect				m_shapeBounds;
	stShapeWithStyle	m_shapeWithStyle;
};

//------------------------------------------------------------------------------

struct stDefineSprite : public stSWFBase
{
	stDefineSprite() : stSWFBase(TagID_DEFINESPRITE){m_tags=NULL;};
	virtual ~stDefineSprite()
	{		
		if (m_tags)
		{
			while (m_tags)
			{
				stSWFTags * tmp = m_tags;
				m_tags = m_tags->m_next;

				delete tmp;
			}
		}
	}

	int					m_characterID;	// ID UI16 ID of sprite 
	int					m_FrameCount;	// UI16 Total number of frames in sprite
	stSWFTags*			m_tags;
};

//------------------------------------------------------------------------------
struct stRemoveObject2 : public stSWFBase
{
	stRemoveObject2() : stSWFBase(TagID_REMOVEOBJECT2){};
	int	m_Depth;						// UI16 Depth of character 
};

//------------------------------------------------------------------------------
struct stPlaceObject2 : public stSWFBase
{
	stPlaceObject2() : stSWFBase(TagID_PLACEOBJECT2){};

	int m_PlaceFlagsReserved;			// UB[2] Reserved Flags (undocumented). These might control clipping (PH 99.04.01)

	int m_PlaceFlagHasName;				// UB[1] Has name 
	int m_PlaceFlagHasRatio;			// UB[1] Has ratio 
	int m_PlaceFlagHasColorTransform;	// UB[1] Has color transform 
	int m_PlaceFlagHasMatrix;			// UB[1] Has matrix 
	int m_PlaceFlagHasCharacter;		// UB[1] Places a character 
	int m_PlaceFlagMove;				// UB[1] Transforms a character that has already been placed.  
	// Typically, an animated object has PlaceFlagHasCharacter set in the first 
	// frame, and PlaceFlagMove set in subsequent frames, with new color or transform matrices. (DM 99.07.13)
	int			m_Depth;				// UI16 Depth of character 
	int			m_CharacterId;			// If PlaceFlagHasCharacter = 1 UI16 ID of character to place 
	stMatrix	m_Matrix;				// If PlaceFlagHasMatrix = 1 MATRIX Transform matrix data 
	stCXForm	m_ColorTransform;		// If PlaceFlagHasColorTransform = 1 CXFORM Color transform data 
	int			m_Ratio;				// If PlaceFlagHasRatio = 1 UI16 Used in tweening. Indicates the amount of target 
	// shape in an intermediate tween shape. Ratio ranges in value between 0 (at the frame 
	// associated with the source shape) and less than 65536 (at the last tween frame) and
	// reflects the amount of easing in and out.  (RH 99.05.14)

	// The ratio field is used in conjunction with the MOVE flag to control 
	// MorphShape's. Each MorphShape has to sets of edges (edges1 and edges2). 
	// When a PlaceObject2 is used on a MorphShape with the MOVE and HasRatio flags 
	// set, the ratio says how far along a linear interpolation between edges1 and 
	// edges2 the shape is. ratio=0 means just display the edges1 shape, ratio=65535 means 
	// just display the edges2 shape.  (CG 99.07.15) 

	char		m_Name[32];				// If PlaceFlagHasName = 1 STRING Name of character 
};

//------------------------------------------------------------------------------

struct stShowFrame : public stSWFBase // Flags the end of the current frame
{
	stShowFrame() : stSWFBase(TagID_SHOWFRAME){};	
};

//------------------------------------------------------------------------------

struct stBackGroundColour : public stSWFBase
{
	stBackGroundColour() : stSWFBase(TagID_SETBACKGROUNDCOLOR){};	

	unsigned char		m_R,
						m_G, 
						m_B;
};

//------------------------------------------------------------------------------

struct stFrameLabel : public stSWFBase
{
	stFrameLabel() : stSWFBase(TagID_FRAMELABEL){};	

	char				m_name[32];
};

//------------------------------------------------------------------------------

struct stDefineBitsJPEG2 : public stSWFBase
{
	stDefineBitsJPEG2() : stSWFBase(TagID_DEFINEBITSJPEG2)
	{
		m_rgba = 0;
		m_width = 0;
		m_height = 0;
		m_jpgData = 0;
	};	
	~stDefineBitsJPEG2()
	{
		if (m_rgba)
		{
			delete[] m_rgba;
			m_rgba=NULL;
		}
	}

	unsigned short		m_characterID;
	char*				m_jpgData;

	char*				m_rgba;
	int					m_width;
	int					m_height;
};

//------------------------------------------------------------------------------

struct stDefineBitsJPEG3 : public stSWFBase
{
	stDefineBitsJPEG3() : stSWFBase(TagID_DEFINEBITSJPEG3)
	{
		m_rgba = 0;
		m_width = 0;
		m_height = 0;
		m_jpgData = 0;
		m_alphaData = 0;
	};
	~stDefineBitsJPEG3()
	{
		if (m_rgba)
		{
			delete[] m_rgba;
			m_rgba=NULL;
		}
	}

	unsigned short		m_characterID;
	unsigned int		m_sizeJpgData; // Also offset to alpha data
	char*				m_jpgData;
	unsigned char*		m_alphaData;   // size is image jpg [width * height]

	char*				m_rgba;
	int					m_width;
	int					m_height;
};

//------------------------------------------------------------------------------

struct stNamedAnchor : public stSWFBase
{
	stNamedAnchor() : stSWFBase(TagID_FRAMELABEL){};

	char				m_name[32];
	// additional 0
	unsigned char		m_namedAnchorFlag;
};

//------------------------------------------------------------------------------

inline stSWFTags* AddTag( stSWFTags** parent )
{
	if (*parent==NULL)
	{
		*parent = new stSWFTags();
		return *parent;
	}
	else
	{
		stSWFTags* tmp = *parent;
		while ( tmp->m_next )
		{
			tmp = tmp->m_next;
		}
		tmp->m_next = new stSWFTags();
		return tmp->m_next;
	}
}

inline stSWFTags* GetTag(stSWFTags* tag, int i)
{
	stSWFTags* tmp = tag;
	int count = 0;
	while ( tmp )
	{
		if (count==i)
		{
			return tmp;
		}
		count++;
		tmp=tmp->m_next;
	}
	return NULL;
}

inline int GetNumTags(stSWFTags* tag)
{
	stSWFTags* tmp = tag;
	int count = 0;
	while ( tmp )
	{
		count++;
		tmp=tmp->m_next;
	}
	return count;
}

inline stSWFTags* FindTag(stSWFTags* parent, int cID)
{
	stSWFTags* tmp = parent;
	while ( tmp )
	{
		switch (tmp->m_id)
		{
			case TagID_DEFINESHAPE:
			{
				stDefineShape* ds = static_cast<stDefineShape*>(tmp->m_base);
				if (ds->m_characterID == cID)
				{
					return tmp;
				}
			}
			break;

			case TagID_DEFINESPRITE:
			{
				stDefineSprite* st = static_cast<stDefineSprite*>(tmp->m_base);
				if (st->m_characterID == cID)
				{
					return tmp;
				}
			}
			break;

			case TagID_DEFINEBITSJPEG2:
			{
				stDefineBitsJPEG2* st = static_cast<stDefineBitsJPEG2*>(tmp->m_base);
				if (st->m_characterID == cID)
				{
					return tmp;
				}
			}
			break;

			case TagID_DEFINEBITSJPEG3:
			{
				stDefineBitsJPEG3* st = static_cast<stDefineBitsJPEG3*>(tmp->m_base);
				if (st->m_characterID == cID)
				{
					return tmp;
				}
			}
			break;
		}

		tmp=tmp->m_next;
	}
	return NULL;
}
//------------------------------------------------------------------------------

struct stSWF
{
	stSWFHeader m_header;
	stSWFTags*	m_tags;

	stSWF()
	{
		m_tags=NULL;
	}
	~stSWF()
	{
		while (m_tags)
		{
			if (m_tags)
			{
				stSWFTags * tmp = m_tags;
				m_tags = m_tags->m_next;

				delete tmp;
			}
		}
	}

	stSWFTags* FindTag(int cID)
	{
		return ::FindTag(m_tags, cID);
	}
};


//------------------------------------------------------------------------------


// Debug output - dump to testfiles
__forceinline void dprintf(const char *fmt, ...);
void DebugRawDataFile(char* filename, char * data, int size);

// Entry Point
int  ReadFlash(char * szFileName, stSWF* swf);

// Better function for reading bits!!!
int  ReadBitData(char * uncompData, int& val, int numBitsToRead, int & bitOffset);

// All the various functions to read in the chunks
int ReadFillStyle(char * uncompData, stFillStyle* fs, enShapeType shapeType);
int  ReadMatrix(char * uncompData, stMatrix* matrix, bool doPad=true);
int  ReadRect(char * uncompData, stRect* rect);
void ReadDataChunks(char * data, int lenData, stSWFTags** parentTag);
void ReadDataTag(int tagID, int tagLength, char * data, stSWFTags * tag);
int  ReadBackGroundColour(char * uncompData, stBackGroundColour* out);
int  ReadFrameLabel(char * uncompData, stFrameLabel* out);
int  ReadDefineBitsJPEG2(char * uncompData, stDefineBitsJPEG2* out, int tagLen);
int  ReadDefineBitsJPEG3(char * uncompData, stDefineBitsJPEG3* out, int tagLen);
int ReadLineStyle(char * uncompData, stLineStyle* ls, enShapeType shapeType);

int ReadShapeRecordType1(char* uncompData, stShapeWithStyle* sws, stShapeRecordType1* out, int& bitOffset, enShapeType shapeType);
int ReadShapeRecordType2(char * uncompData, stShapeRecordType2* out, int &  bitOffset);

int ReadShapeRecords(char * uncompData, stShapeWithStyle* out, enShapeType shapeType);
int ReadShapeWithStyle(char * uncompData, stShapeWithStyle* out, enShapeType shapeType);
int  ReadDefineShape(char * uncompData, stDefineShape* out, int tagLen);
int  ReadPlaceObject2(char * uncompData, stPlaceObject2* out, int tagLen);
int  ReadDefineSprite(char * uncompData, stDefineSprite* out, int tagLen);
int ReadGradient(char * uncompData, stGradient* out);

int ReadFillStyleArray(char * uncompData, stFillStyleArray* out, enShapeType shapeType);
int ReadLineStyleArray(char * uncompData, stLineStyleArray* out, enShapeType shapeType);
//------------------------------------------------------------------------------


#endif //SWF_H
























