/******************************************************************************/
/*                                                                            */
/*  File: swfdraw.cpp                                                         */
/*  Author: bkenwright@xbdev.net                                              */
/*  Date: 19-01-06                                                            */
/*                                                                            */
/******************************************************************************/

#include <d3dx8.h>
#include "swf.h"
#include "dxdebug.h"

#include "swfdraw.h"
#include "swftextures.h"


//#define DEBUG_SINGLE_STEP
static bool g_doFrame = true;

stSWF*		SWFSpriteObj::m_swf =	NULL;
int			SWFSpriteObj::s_obj	=	0;
SWFTextures	SWFSpriteObj::s_texlist[MAX_TEXTURES];
int			SWFSpriteObj::s_numtex = 0;


/******************************************************************************/
/*                                                                            */
/*  Name: KeyBoard()                                                          */
/*  Date: 22-12-06                                                            */
/*                                                                            */
/*  Used to accept keyboard input and alter some global variables - used      */
/*  for simple analysis of our code - turning wire frame rendering on/off     */
/*  adjusting variables which are used in the generation of our shape etc.    */
/*                                                                            */
/******************************************************************************/
void KeyBoard()
{
	static float oldTime = (float)GetTickCount();
	float timeNow = (float)GetTickCount()+1.0f;

	if (timeNow < oldTime+100)
	{
		return;
	}

	oldTime=timeNow;

	if( GetKeyState(VK_DOWN)	& 0x80 )  g_doFrame=true;

}//End KeyBoard()


//------------------------------------------------------------------------------

SWFDraw::SWFDraw()
{
	m_swf=NULL;
}

//------------------------------------------------------------------------------

bool SWFDraw::Create(IDirect3DDevice8* pD3DDevice, char* szFileName)
{
	Release();

	dprintf("Opening SWF File: %s\n", szFileName);

	m_swf = new stSWF();
	ReadFlash(szFileName, m_swf);

	LoadAllTextures(pD3DDevice);

	D3DXMATRIX mat;
	D3DXMatrixIdentity(&mat);
	stCXForm cxForm;

	m_obj.m_swf = m_swf;
	m_obj.Create(0, m_swf->m_tags, mat, -1, cxForm);
	return true;
}

//------------------------------------------------------------------------------

bool SWFDraw::LoadAllTextures(IDirect3DDevice8* pD3DDevice)
{	
	int numTags = GetNumTags(m_swf->m_tags);
	for (int i=0; i<numTags; i++)
	{
		stSWFTags* tag = GetTag(m_swf->m_tags, i);

		if (tag==NULL)
		{
			continue;
		}


		switch (tag->m_id)
		{
			case TagID_DEFINEBITSJPEG2:
			case TagID_DEFINEBITSJPEG3:
			{
			}
			break;
			
			default:
			{
				continue;
			}
		}


		if (SWFSpriteObj::s_numtex >= MAX_TEXTURES-1)
		{
			OutputDebugString("***ERROR"""" Array max size 500\n");
			_asm { int 0x10 }
			return false;
		}

		int indx = SWFSpriteObj::s_numtex;
		SWFSpriteObj::s_numtex++;
		SWFTextures* tex = &SWFSpriteObj::s_texlist[indx];

		int iw = 0;
		int ih = 0;
		int cID = 0;
		char* rgba = NULL;

		switch (tag->m_id)
		{
			case TagID_DEFINEBITSJPEG2:
			{
				stDefineBitsJPEG2* jpg = static_cast<stDefineBitsJPEG2*>(tag->m_base);
				rgba = jpg->m_rgba;
				iw   = jpg->m_width;
				ih   = jpg->m_height;
				cID  = jpg->m_characterID;
			}
			break;

			case TagID_DEFINEBITSJPEG3:
			{
				stDefineBitsJPEG3* jpg = static_cast<stDefineBitsJPEG3*>(tag->m_base);
				rgba = jpg->m_rgba;
				iw   = jpg->m_width;
				ih   = jpg->m_height;
				cID  = jpg->m_characterID;
			}
			break;
		}
		tex->Create(pD3DDevice, rgba, iw, ih, cID);
	}

	return true;
}//End LoadAllTextures()

//------------------------------------------------------------------------------

bool SWFDraw::Release()
{
	m_obj.Release();

	if (m_swf)
	{
		delete m_swf;
		m_swf=NULL;
	}


	for (int i=0; i<SWFSpriteObj::s_numtex; i++)
	{
		SWFTextures* tex = &SWFSpriteObj::s_texlist[i];
		tex->Release();
	}
	SWFSpriteObj::s_numtex=0;

	return true;
}


//------------------------------------------------------------------------------
int SWFDraw::FindFrame(char* labelname)
{
	if ((m_swf==NULL) || (labelname==NULL))
	{
		return 0;
	}

	int numTags = GetNumTags(m_swf->m_tags);
	for (int i=0; i<numTags; i++)
	{
		stSWFTags* tag = GetTag(m_swf->m_tags, i);
		if (tag->m_id==TagID_FRAMELABEL)
		{
			stFrameLabel* fl = static_cast<stFrameLabel*>(tag->m_base);
			
			if (strcmp(fl->m_name, labelname)==NULL)
			{
				return i;
			}
		}
	}

	return -1;
}

//------------------------------------------------------------------------------

bool SWFDraw::Select(IDirect3DDevice8* pD3DDevice, char* startLabel, char* endLabel, bool loop)
{
	if (m_swf==NULL) return false;

	int startIndx = FindFrame(startLabel);
	int endIndx   = FindFrame(endLabel);

	int numTags = GetNumTags(m_swf->m_tags);

	if ( startIndx<0 || startIndx>numTags ) return false;
	if ( endIndx<0   || endIndx>numTags )   return false;

	m_obj.Select(pD3DDevice, startIndx, endIndx, loop);

	return true;
}


void SWFDraw::Render(IDirect3DDevice8* pD3DDevice)
{
	if (m_swf==NULL) return;

	KeyBoard();

	extern HWND g_hWnd;
	RECT wr;
	GetClientRect(g_hWnd, &wr);

	stRect* r = &m_swf->m_header.m_FrameSize;

	float w0 = ((r->m_Xmax - r->m_Xmin)/20.0f);
	float h0 = ((r->m_Ymax - r->m_Ymin)/20.0f);
	float w1 = (float)(wr.right - wr.left);
	float h1 = (float)(wr.bottom - wr.top);

	float left = wr.left + (w1-w0)/2.0f;
	float top  = wr.top  + (h1-h0)/2.0f;
	

	D3DXMATRIX mat;
	D3DXMatrixTranslation(&mat, left, top, 0);

	

	D3DXVECTOR3 topLeft = D3DXVECTOR3((float)r->m_Xmin/20.0f, (float)r->m_Xmin/20.0f, 0);
	D3DXVec3TransformCoord(&topLeft, &topLeft, &mat);

	DebugWireDrawBox2D(pD3DDevice, topLeft.x, topLeft.y, w0, h0);


	stCXForm tc;
	
	m_obj.Render(pD3DDevice, &mat, 0, &tc);

	#ifdef DEBUG_SINGLE_STEP
	if (g_doFrame)
	{
		g_doFrame=false;
		m_obj.Update(pD3DDevice);
	}
	#else

	static float oldTime = (float)GetTickCount();
	float timeNow = (float)GetTickCount()+1.0f;
	if (timeNow < oldTime+30)
	{
		return;
	}
	oldTime=timeNow;

	m_obj.Update(pD3DDevice);
	#endif
}


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

SWFSpriteObj::SWFSpriteObj()
{
	m_cID = -1;
	D3DXMatrixIdentity(&m_mat);
	m_tags=NULL;
	m_frame=0;
	m_active=false;
	m_depth=-2;
	for (int i=0; i<MAX_CHILD; i++)
	{
		m_child[i]=NULL;
	}
	m_texID=-1;
}
SWFSpriteObj::~SWFSpriteObj()
{
	Release();
}

//------------------------------------------------------------------------------

void SWFSpriteObj::Select(IDirect3DDevice8* pD3DDevice, int startIndx, int endIndx, bool loop)
{
	if (m_swf==NULL) return;

	if ( startIndx<0 || startIndx>m_numTags ) return;
	if ( endIndx<0   || endIndx>m_numTags )   return;

	m_startIndx	= startIndx;
	m_endIndx	= endIndx;
	m_loop		= loop;
	m_frame		= 0;
	m_donePlay  = false;

	Release();

	while (m_frame < m_startIndx )
	{
		ShowFrame(pD3DDevice);
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::Create(int cID, stSWFTags* tag, D3DXMATRIX mat, int depth, stCXForm tc)
{
	m_cID=cID;
	m_tags=tag;
	m_mat=mat;
	m_depth=depth;
	m_active=true;
	m_numVerts=0;
	m_draw=false;
	s_obj++;

	if (s_obj >= MAX_CHILD )
	{
		OutputDebugString("***ERROR"""" Array max size 500\n");
		_asm { int 0x10 }
	}

	m_numTags = GetNumTags(m_tags);
	m_startIndx=0;
	m_endIndx=m_numTags;
	m_loop=true;
	m_donePlay=false;
	m_cxForm=tc;
}
//------------------------------------------------------------------------------

void SWFSpriteObj::ReleaseAllObjects()
{
	for (int i=0; i<MAX_CHILD; i++)
	{
		if (m_child[i])
		{
			m_child[i]->Release();
			delete m_child[i];
			m_child[i]=NULL;
		}
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::Release()
{
	s_obj--;
	m_cID=-2;
	m_active=false;
	m_texID=-1;

	ReleaseAllObjects();
}

//------------------------------------------------------------------------------

void SWFSpriteObj::Update(IDirect3DDevice8* pD3DDevice)
{
	if (m_frame >= m_endIndx)
	{
		m_donePlay=true;
		if (m_loop)
		{
			Select(pD3DDevice, m_startIndx, m_endIndx, true);
		}
		else
		{
			return;
		}
	}

	ShowFrame(pD3DDevice);
}

//------------------------------------------------------------------------------

void SWFSpriteObj::ShowFrame(IDirect3DDevice8* pD3DDevice)
{
	for (int i=0; i<MAX_CHILD; i++)
	{
		if (m_child[i])
		{
			m_child[i]->ShowFrame(pD3DDevice);
		}
	}

	for (int i=m_frame; i<m_numTags; i++)
	{
		stSWFTags* tag = GetTag(m_tags, i);
		if (tag)
		{
			switch(tag->m_id)
			{
				case TagID_PLACEOBJECT2:
				case TagID_REMOVEOBJECT2:
				{
					DoTag(pD3DDevice, NULL, tag, -1, NULL);
				}
				break;

				case TagID_SHOWFRAME:
				case TagID_END:
				{
					m_frame = i+1;
					return;
				}
				break;
			}
		}
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::Render(IDirect3DDevice8* pD3DDevice, D3DXMATRIX* pmat, int depth, stCXForm* tc)
{
	D3DXVECTOR2 maxverts[20];

	D3DXMATRIX mat = m_mat * (*pmat);

	stCXForm cxForm = m_cxForm * (*tc);

	for (int j=0; j<MAX_CHILD; j++)
	{
		SWFSpriteObj* obj = m_child[j];
		if (obj)
		{
			if (obj->m_draw==false)
			{
				obj->Render(pD3DDevice, &mat, obj->m_depth, &cxForm);
			}
			else
			{
				if (obj->m_active)
				{
					for (int i=0; i<obj->m_numVerts; i++)
					{
						D3DXMATRIX mm =  obj->m_mat * mat;
						D3DXVECTOR3 tmp;

						tmp.x=obj->m_verts[i].x;
						tmp.y=obj->m_verts[i].y;
						tmp.z=depth * 20.0f;
						D3DXVec3TransformCoord(&tmp, &tmp, &mm);

						maxverts[i].x=tmp.x;
						maxverts[i].y=tmp.y;
					}

					if (obj->m_texID >= -1)
					{

						float mulr=1.0f;
						float mulg=1.0f;
						float mulb=1.0f;
						float mula=1.0f;

						int addr=0;
						int addg=0;
						int addb=0;
						int adda=0;

						mulr = cxForm.m_RedMultTerm;
						mulg = cxForm.m_GreenMultTerm;
						mulb = cxForm.m_BlueMultTerm;
						mula = cxForm.m_AlphaMultTerm;

						addr = cxForm.m_RedAddTerm;
						addg = cxForm.m_GreenAddTerm;
						addb = cxForm.m_BlueAddTerm;
						adda = cxForm.m_AlphaAddTerm;

						IDirect3DTexture8* pTexture = NULL;
						int numTex = SWFSpriteObj::s_numtex;
						for (int i=0; i<numTex; i++)
						{
							SWFTextures* tex = &SWFSpriteObj::s_texlist[i];
							if (tex->m_cID==obj->m_texID)
							{
								pTexture = tex->m_texture;
								break;
							}
						}

						DebugDrawTexList2D(pD3DDevice, pTexture, &maxverts[0], obj->m_texcoords, obj->m_numVerts, depth,
										   mulr, mulg, mulb, mula,
										   addr, addg, addb, adda);
					}
					else
					{
						DebugDrawLineList2D(pD3DDevice, &maxverts[0], obj->m_numVerts);
					}
				}
			}
		}
	}
}

//------------------------------------------------------------------------------


void SWFSpriteObj::DoTag(IDirect3DDevice8* pD3DDevice, D3DXMATRIX* mat, stSWFTags* tag, int depth, stCXForm* tc)
{
	switch (tag->m_id)
	{
		case TagID_DEFINESHAPE:
		{
			stDefineShape* ds = static_cast<stDefineShape*>(tag->m_base);
			DoDefineShape(pD3DDevice, mat, ds, depth, tc);
		}
		break;

		case TagID_DEFINESPRITE:
		{
			stDefineSprite* ds = static_cast<stDefineSprite*>(tag->m_base);
			DoDefineSprite(pD3DDevice, mat, ds, depth, tc);
		}
		break;

		case TagID_PLACEOBJECT2:
		{
			stPlaceObject2* po = static_cast<stPlaceObject2*>(tag->m_base);
			DoPlaceObject2(pD3DDevice, po);
		}
		break;

		case TagID_REMOVEOBJECT2:
		{
			stRemoveObject2* ro = static_cast<stRemoveObject2*>(tag->m_base);
			DoRemoveObject2(pD3DDevice, ro);
		}
		break;

		default:
		{
			int jj = 2;
		}
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::DoDefineShape(IDirect3DDevice8* pD3DDevice, D3DXMATRIX* mat, stDefineShape* ds, int depth,  stCXForm* tc)
{
	//stRect* r = &ds->m_shapeBounds;
	//DebugWireDrawBox2D(pD3DDevice, (float)r->m_Xmin/20.0f, (float)r->m_Ymin/20.0f, (float)r->m_Xmax/20.0f, (float)r->m_Ymax/20.0f);
	
	// Temp max numb poitns
	int numPoints = 0;
	D3DXVECTOR2 points[20];
	D3DXVECTOR2 tex[20];

	stRect* bounds = &ds->m_shapeBounds;
	float dx = (float)(bounds->m_Xmax - bounds->m_Xmin)/20.0f;
	float dy = (float)(bounds->m_Ymax - bounds->m_Ymin)/20.0f;

	stSWFTags* imgTag = NULL;
	D3DXVECTOR2 imgTrans;

	float itx = 0;
	float ity = 0;

	stShapeWithStyle* sws = &ds->m_shapeWithStyle;
	stShapeRecordBase* sr = sws->m_ShapeRecord;
	while (sr)
	{
		switch (sr->m_shapeRecordType)
		{
			case ShapeRecord1: // Line Style, Fill Style and Positions
			{
				stShapeRecordType1* sr1 = static_cast<stShapeRecordType1*>(sr);
				points[numPoints] = D3DXVECTOR2(0,0);
				if (sr1->m_StateMoveTo)
				{
					points[numPoints].x=(float)sr1->m_MoveX/20.0f;
					points[numPoints].y=(float)sr1->m_MoveY/20.0f;
				}

				int indx = -1;

				if (sr1->m_StateFillStyle1)
				{
					indx = sr1->m_Fill1Style-1; // goes from 1 to val, not 0
				}
				else if (sr1->m_StateFillStyle0)
				{
					indx = sr1->m_Fill0Style-1; ; // goes from 1 to val, not 0
				}

				if ( indx >= 0 )
				{
					stFillStyleArray* fsa = &sws->m_FillStyleArray;

					stFillStyle* fs = &fsa->m_FillStyles[indx];

					if ((fs->m_FillStyleType==FS_TILED) || (fs->m_FillStyleType==FS_CLIPPED))
					{
						imgTag = m_swf->FindTag(fs->m_BitmapID);

						imgTrans.x = (float)fs->m_BitmapMatrix.m_xTranslate/20.0f; /// to fix...negative values...
						imgTrans.y = (float)fs->m_BitmapMatrix.m_yTranslate/20.0f;


						float iw = 0.0f;
						float ih = 0.0f;
						switch (imgTag->m_id)
						{
							case TagID_DEFINEBITSJPEG2:
							{
								stDefineBitsJPEG2* jpg = static_cast<stDefineBitsJPEG2*>(imgTag->m_base);
								iw  = (float)jpg->m_width;
								ih  = (float)jpg->m_height;
							}
							break;

							case TagID_DEFINEBITSJPEG3:
							{
								stDefineBitsJPEG3* jpg = static_cast<stDefineBitsJPEG3*>(imgTag->m_base);
								iw  = (float)jpg->m_width;
								ih  = (float)jpg->m_height;
							}
							break;

						}

						float isx = 0.0f;
						float isy = 0.0f;
						if (fs->m_BitmapMatrix.m_hasScale)
						{
							isx = fs->m_BitmapMatrix.m_xScale;
							isy = fs->m_BitmapMatrix.m_yScale;
						}

						itx = (iw*isx)/20.0f;
						ity = (ih*isy)/20.0f;

						float px = (points[numPoints].x - imgTrans.x);
						float py = (points[numPoints].y - imgTrans.y);

						tex[numPoints].x = px/itx;
						tex[numPoints].y = py/ity;

					}
				}				

				numPoints++;
			}
			break;

			case ShapeRecord2: // Straight or curved edge 
			{
				stShapeRecordType2* sr2 = static_cast<stShapeRecordType2*>(sr);
				int prev = numPoints<=0? 0:(numPoints-1);

				if (sr2->m_EdgeFlag==0)
				{
					EdgeRecordType0* edge0 = &sr2->m_EdgeRecord0;

					points[numPoints].x = points[prev].x + edge0->m_ControlDeltaX/20.0f;
					points[numPoints].y = points[prev].y + edge0->m_ControlDeltaY/20.0f;
				}
				else
				{
					EdgeRecordType1* edge1 = &sr2->m_EdgeRecord1;
					points[numPoints]=points[prev];
					if (edge1->m_VertLineFlag)
					{
						points[numPoints].y += (float)edge1->m_VLineDeltaY/20.0f;
					}
					else
					{
						points[numPoints].x += (float)edge1->m_VLineDeltaX/20.0f;
					}
					
				}

				if (imgTag)
				{
					float px = (points[numPoints].x - imgTrans.x);
					float py = (points[numPoints].y - imgTrans.y);

					tex[numPoints].x = px/itx;
					tex[numPoints].y = py/ity;
				}

				numPoints++;
			}
			break;

			case ShapeRecordEnd:
			{
			}
			break;
		}

		sr = sr->m_next;
	}

	if (numPoints>1)
	{

		SWFSpriteObj* obj = NULL;
		for (int i=0; i<MAX_CHILD; i++)
		{
			if (m_child[i]==NULL)
			{
				obj = new SWFSpriteObj();
				m_child[i] = obj;
				break;
			}
		}


		for (int i=0; i<numPoints; i++)
		{
			obj->m_verts[i].x = points[i].x;
			obj->m_verts[i].y = points[i].y;

			obj->m_texcoords[i].x   = tex[i].x;
			obj->m_texcoords[i].y   = tex[i].y;
		}

		if (imgTag)
		{
			
			int iw = 0;
			int ih = 0;
			char* rgba = NULL;
			int cID = 0;

			switch (imgTag->m_id)
			{
				case TagID_DEFINEBITSJPEG2:
				{
					stDefineBitsJPEG2* jpg = static_cast<stDefineBitsJPEG2*>(imgTag->m_base);
					rgba = jpg->m_rgba;
					iw   = jpg->m_width;
					ih   = jpg->m_height;
					cID  = jpg->m_characterID;
				}
				break;

				case TagID_DEFINEBITSJPEG3:
				{
					stDefineBitsJPEG3* jpg = static_cast<stDefineBitsJPEG3*>(imgTag->m_base);
					rgba = jpg->m_rgba;
					iw   = jpg->m_width;
					ih   = jpg->m_height;
					cID  = jpg->m_characterID;
				}
				break;

			}
			/*
			HRESULT hr = 
			pD3DDevice->CreateTexture(	iw, 
										ih, 
										1,
										//D3DUSAGE_RENDERTARGET, 
										D3DUSAGE_DYNAMIC, // Usage flag
										D3DFMT_A8R8G8B8,
										D3DPOOL_DEFAULT, 
										//D3DPOOL_SYSTEMMEM,
										&obj->m_texture );

			IDirect3DTexture8* texture = obj->m_texture;
			D3DSURFACE_DESC desc;
			texture->GetLevelDesc (0, &desc);

			D3DLOCKED_RECT rect;
			hr = 
			texture->LockRect (0, &rect, 0, D3DLOCK_NOSYSLOCK);
			

			for (int y=0; y< (int)desc.Height; y++)
			{
				for(int x =0; x < (int)desc.Width; x++)
				{
					DWORD* pixel = ((DWORD*) (((BYTE*) rect.pBits) +rect. Pitch*y+x*sizeof (DWORD)));

					int i = (x + (iw)*y) * 4;

					
					unsigned char cr = rgba[i + 3];
					unsigned char cg = rgba[i + 2];
					unsigned char cb = rgba[i + 1];
					unsigned char ca = rgba[i + 0];			

					DWORD colour = (ca<<24)|(cr<<16)|(cg<<8)|(cb<<0);

					*pixel = colour;

				}
			}

			texture->UnlockRect (0);
			*/

			obj->m_texID = cID;
		}

		obj->m_cxForm = *tc;

		obj->m_depth = depth;
		obj->m_cID = ds->m_characterID;
		obj->m_mat = *mat;

		obj->m_draw=true;
		obj->m_active=true;
		obj->m_numVerts = numPoints;
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::DoDefineSprite(IDirect3DDevice8* pD3DDevice, D3DXMATRIX* mat, stDefineSprite* ds, int depth,  stCXForm* tc)
{
	for (int i=0; i<MAX_CHILD; i++)
	{
		if (m_child[i]==NULL)
		{
			m_child[i] = new SWFSpriteObj();
			m_child[i]->Create(ds->m_characterID, ds->m_tags, *mat, depth, *tc);
			m_child[i]->ShowFrame(pD3DDevice);
			break;
		}
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::DoPlaceObject2(IDirect3DDevice8* pD3DDevice, stPlaceObject2* ds)
{
	D3DXMATRIX matx;
	D3DXMatrixIdentity(&matx);

	if (ds->m_PlaceFlagHasMatrix)
	{
		stMatrix* fm = &ds->m_Matrix;

		if (fm->m_hasScale)
		{
			matx._11 = fm->m_xScale;
			matx._22 = fm->m_yScale;
		}

		if (fm->m_hasRotate)
		{
			float aa = asinf(fm->m_RotateSkew0);
			matx._12 = fm->m_RotateSkew0;
			matx._21 = fm->m_RotateSkew1;
		}

		matx._41 =  fm->m_xTranslate/20.0f;
		matx._42 =  fm->m_yTranslate/20.0f;
	}

	stCXForm tc;
	if (ds->m_PlaceFlagHasColorTransform)
	{
		tc = tc * ds->m_ColorTransform;
	}


	if (ds->m_PlaceFlagHasCharacter==1 && 
		ds->m_PlaceFlagMove==0)
	{
		stSWFTags* tag = m_swf->FindTag(ds->m_CharacterId);

		if (tag)
		{
			DoTag(pD3DDevice, &matx, tag, ds->m_Depth, &tc);
		}
	}

	if (ds->m_PlaceFlagHasCharacter==0 &&
		ds->m_PlaceFlagMove==1 )
		//ds->m_PlaceFlagHasMatrix==1)
	{
		for (int i=0; i<MAX_CHILD; i++)
		{
			if (m_child[i])
			{
				if (m_child[i]->m_depth==ds->m_Depth)
				{
					if (ds->m_PlaceFlagHasMatrix==1)
					{
						m_child[i]->m_mat = matx;
					}

					if (ds->m_PlaceFlagHasColorTransform)
					{
						m_child[i]->m_cxForm = tc;
					}
				}
			}
		}
	}

	if (ds->m_PlaceFlagHasCharacter==1 &&
		ds->m_PlaceFlagMove==1)
	{
		// The character at the specified Depth is removed, and a new character 
		// (with ID of CharacterId) is placed at that depth. Other fields set the 
		// attributes of this new character.

		dprintf(" ");
	}
}

//------------------------------------------------------------------------------

void SWFSpriteObj::DoRemoveObject2(IDirect3DDevice8* pD3DDevice, stRemoveObject2* ds)
{
	int depth = ds->m_Depth;

	for (int i=0; i<MAX_CHILD; i++)
	{
		if (m_child[i])
		{
			if (m_child[i]->m_depth==depth)
			{
				m_child[i]->Release();
				delete m_child[i];
				m_child[i]=NULL;
			}
		}
	}
}

//------------------------------------------------------------------------------









